# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Pango

from lollypop.define import App


class Placeholder(Gtk.Bin):
    def __init__(self, text, icon_name):
        """
            Init placeholder
            @param text as str
            @param icon_name as str
        """
        Gtk.Bin.__init__(self)
        self.__label = Gtk.Label.new()
        self.__label.show()
        self.__label.set_markup("%s" % GLib.markup_escape_text(text))
        self.__label.set_line_wrap_mode(Pango.WrapMode.WORD)
        self.__label.set_line_wrap(True)
        self.set_folded(App().window.folded)
        label_style = self.__label.get_style_context()
        label_style.add_class("dim-label")
        image = Gtk.Image.new_from_icon_name(icon_name,
                                             Gtk.IconSize.DIALOG)
        image.show()
        image.get_style_context().add_class("dim-label")
        grid = Gtk.Grid()
        grid.show()
        grid.set_margin_start(20)
        grid.set_margin_end(20)
        grid.set_column_spacing(20)
        grid.add(image)
        grid.add(self.__label)
        grid.set_vexpand(True)
        grid.set_hexpand(True)
        grid.set_property("halign", Gtk.Align.CENTER)
        grid.set_property("valign", Gtk.Align.CENTER)
        self.add(grid)

    def set_folded(self, folded):
        """
            Set label size
            @param folded as bool
        """
        style_context = self.__label.get_style_context()
        if folded:
            style_context.remove_class("text-xx-large")
            style_context.add_class("text-x-large")
        else:
            style_context.remove_class("text-x-large")
            style_context.add_class("text-xx-large")



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio

from random import shuffle

from lollypop.objects_album import Album
from lollypop.objects_track import Track
from lollypop.logger import Logger
from lollypop.define import App, Repeat, StorageType
from lollypop.utils import sql_escape, get_network_available
from lollypop.utils import get_default_storage_type, emit_signal
from lollypop.utils_album import tracks_to_albums


class AutoSimilarPlayer:
    """
        Manage playback when going to end
    """

    def __init__(self):
        """
            Init player
        """
        self.__next_cancellable = Gio.Cancellable()
        self.__radio_cancellable = Gio.Cancellable()
        self.connect("next-changed", self.__on_next_changed)

    def next_album(self):
        """
            Get next album to add
            @return Album
        """
        genre_ids = App().artists.get_genre_ids(self.current_track.artist_ids,
                                                StorageType.COLLECTION)
        track_ids = App().tracks.get_randoms(genre_ids,
                                             StorageType.COLLECTION,
                                             1,
                                             100)
        if track_ids:
            return Track(track_ids[0]).album
        return None

    def play_radio_from_collection(self, artist_ids):
        """
            Play a radio from collection for artist ids
            @param artist_ids as [int]
        """
        genre_ids = App().artists.get_genre_ids(artist_ids,
                                                StorageType.COLLECTION)
        track_ids = App().tracks.get_randoms(genre_ids,
                                             StorageType.COLLECTION,
                                             False,
                                             100)
        albums = tracks_to_albums(
            [Track(track_id) for track_id in track_ids], False)
        self.play_albums(albums)

    def play_radio_from_spotify(self, artist_ids):
        """
            Play a radio from the Spotify for artist ids
            @param artist_ids as [int]
        """
        self.__play_radio_common()
        if get_network_available("SPOTIFY") and\
                get_network_available("YOUTUBE"):
            from lollypop.similars_spotify import SpotifySimilars
            similars = SpotifySimilars()
            self.__load_similars(similars, artist_ids)

    def play_radio_from_lastfm(self, artist_ids):
        """
            Play a radio from the Last.fm for artist ids
            @param artist_ids as [int]
        """
        self.__play_radio_common()
        if get_network_available("LASTFM") and\
                get_network_available("YOUTUBE"):
            from lollypop.similars_lastfm import LastFMSimilars
            similars = LastFMSimilars()
            self.__load_similars(similars, artist_ids)

    def play_radio_from_deezer(self, artist_ids):
        """
            Play a radio from the Last.fm for artist ids
            @param artist_ids as [int]
        """
        self.__play_radio_common()
        if get_network_available("DEEZER") and\
                get_network_available("YOUTUBE"):
            from lollypop.similars_deezer import DeezerSimilars
            similars = DeezerSimilars()
            self.__load_similars(similars, artist_ids)

    def play_radio_from_loved(self, artist_ids):
        """
            Play a radio from artists loved tracks
            @param artist_ids as [int]
        """
        track_ids = App().tracks.get_loved_track_ids(artist_ids,
                                                     StorageType.ALL)
        shuffle(track_ids)
        albums = tracks_to_albums([Track(track_id) for track_id in track_ids])
        App().player.play_albums(albums)

    def play_radio_from_populars(self, artist_ids):
        """
            Play a radio from artists popular tracks
            @param artist_ids as [int]
        """
        track_ids = App().tracks.get_populars(artist_ids, StorageType.ALL,
                                              False, 100)
        shuffle(track_ids)
        albums = tracks_to_albums([Track(track_id) for track_id in track_ids])
        App().player.play_albums(albums)

    @property
    def radio_cancellable(self):
        """
            Get cancellable
            @return Gio.Cancellable
        """
        return self.__radio_cancellable

#######################
# PROTECTED           #
#######################
    def _on_stream_start(self, bus, message):
        """
            Cancel radio loading if current not a web track
            @param bus as Gst.Bus
            @param message as Gst.Message
        """
        if not self.current_track.is_web:
            self.__radio_cancellable.cancel()

#######################
# PRIVATE             #
#######################
    def __load_similars(self, similars, artist_ids):
        """
            Load similars for artist ids
            @param similars as Similars
            @param artist ids as [int]
        """
        similars.connect("match-track", self.__on_match_track)
        similars.connect("finished", self.__on_finished)
        self.clear_albums()
        App().task_helper.run(similars.load_similars,
                              artist_ids,
                              StorageType.EPHEMERAL,
                              self.__radio_cancellable)

    def __play_radio_common(self):
        """
            Emit signal and reset cancellable
        """
        emit_signal(self, "loading-changed", True, Track())
        self.__radio_cancellable.cancel()
        self.__radio_cancellable = Gio.Cancellable()

    def __get_album_from_artists(self,  similar_artist_ids):
        """
            Add a new album to playback
            @param similar_artist_ids as [int]
            @return Album
        """
        # Get an album
        storage_type = get_default_storage_type()
        album_ids = App().albums.get_ids(
            [], similar_artist_ids, storage_type, False)
        shuffle(album_ids)
        while album_ids:
            album_id = album_ids.pop(0)
            if album_id not in self.album_ids:
                return Album(album_id, [], [], False)
        return None

    def __get_artist_ids(self, artists):
        """
            Get valid artist ids from list
            @param artists as []
            @return [int]
        """
        similar_artist_ids = []
        for (artist, cover_uri) in artists:
            if self.__next_cancellable.is_cancelled():
                return []
            similar_artist_id = App().artists.get_id_for_escaped_string(
                sql_escape(artist.lower()))
            if similar_artist_id is not None:
                if App().artists.has_albums(similar_artist_id):
                    similar_artist_ids.append(similar_artist_id)
        return similar_artist_ids

    def __on_get_artist_ids(self, similar_artist_ids, remote):
        """
            Get one albums from artist ids
            @param similar_artist_ids as [int]
            @param remote as bool
        """
        if self.__next_cancellable.is_cancelled():
            return
        album = None
        if similar_artist_ids:
            album = self.__get_album_from_artists(similar_artist_ids)
        if album is not None:
            Logger.info("Found a similar album")
            self.add_album(album)
        elif remote:
            from lollypop.similars_local import LocalSimilars
            similars = LocalSimilars()
            App().task_helper.run(
                similars.get_similar_artists,
                App().player.current_track.artist_ids,
                self.__next_cancellable,
                callback=(self.__on_get_local_similar_artists,))

    def __on_get_local_similar_artists(self, artists):
        """
            Add one album from artists to player
            @param artists as []
        """
        if self.__next_cancellable.is_cancelled():
            return
        App().task_helper.run(self.__get_artist_ids, artists,
                              callback=(self.__on_get_artist_ids, False))

    def __on_get_similar_artists(self, artists):
        """
            Add one album from artists to player
            @param artists as []
        """
        if self.__next_cancellable.is_cancelled():
            return
        App().task_helper.run(self.__get_artist_ids, artists,
                              callback=(self.__on_get_artist_ids, True))

    def __on_next_changed(self, player):
        """
            Add a new album if playback finished and wanted by user
        """
        self.__next_cancellable.cancel()
        # Do not load an album if a radio is loading
        if not self.__radio_cancellable.is_cancelled() or not self._albums:
            return
        self.__next_cancellable = Gio.Cancellable()
        # Check if we need to add a new album
        if App().settings.get_enum("repeat") == Repeat.AUTO_SIMILAR and\
                player.next_track.id is None and\
                player.current_track.id is not None and\
                player.current_track.id >= 0 and\
                player.current_track.artist_ids:
            from lollypop.similars import Similars
            similars = Similars()
            App().task_helper.run(
                similars.get_similar_artists,
                player.current_track.artist_ids,
                self.__next_cancellable,
                callback=(self.__on_get_similar_artists,))

    def __on_match_track(self, similars, track_id, storage_type):
        """
            Load/Play track album
            @param similars as Similars
            @param track_id as int
            @param storage_type as StorageType
        """
        track = Track(track_id)
        album = track.album
        if self.albums:
            self.add_album(album)
        else:
            self.play_album(album)

    def __on_finished(self, similars):
        """
            Cancel radio loading
            @param similars as Similars
        """
        self.__radio_cancellable.cancel()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, GLib, Gio, Gst

from datetime import datetime
from gettext import gettext as _

from lollypop.define import App, ArtSize, ArtBehaviour
from lollypop.define import MARGIN_BIG
from lollypop.widgets_player_progress import ProgressPlayerWidget
from lollypop.widgets_player_buttons import ButtonsPlayerWidget
from lollypop.widgets_player_artwork import ArtworkPlayerWidget
from lollypop.widgets_player_label import LabelPlayerWidget
from lollypop.container import Container
from lollypop.logger import Logger
from lollypop.helper_signals import SignalsHelper, signals_map


class FullScreen(Gtk.Window, SignalsHelper):
    """
        Show a fullscreen window showing current track context
    """

    @signals_map
    def __init__(self):
        """
            Init window
        """
        Gtk.Window.__init__(self)
        return [
                (App().player, "current-changed", "_on_current_changed")
        ]

    def delayed_init(self):
        """
            Delay real init to get App().window == self
        """
        self.get_style_context().add_class("black")
        self.set_title("Lollypop")
        self.__allocation = Gdk.Rectangle()
        self.set_application(App())
        self.__timeout_id = None
        self.__signal1_id = self.__signal2_id = None
        self.__background_id = None
        self.set_decorated(False)
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/FullScreen.ui")
        builder.connect_signals(self)
        self.__progress_widget = ProgressPlayerWidget(True)
        self.__progress_widget.show()
        self.__progress_widget.set_property("halign", Gtk.Align.CENTER)
        self.__progress_widget.set_size_request(500, -1)
        self.__buttons_widget = ButtonsPlayerWidget(["banner-button"])
        self.__buttons_widget.show()
        self.__buttons_widget.set_size_request(500, -1)
        self.__buttons_widget.set_property("valign", Gtk.Align.CENTER)
        self.__buttons_widget.set_property("halign", Gtk.Align.CENTER)
        self.__artwork_widget = ArtworkPlayerWidget(
            ArtBehaviour.CACHE | ArtBehaviour.CROP_SQUARE)
        self.__artwork_widget.show()
        self.__artwork_widget.set_vexpand(True)
        self.__artwork_widget.set_art_size(ArtSize.FULLSCREEN,
                                           ArtSize.FULLSCREEN)
        self.__artwork_widget.set_property("valign", Gtk.Align.CENTER)
        self.__artwork_widget.set_property("halign", Gtk.Align.CENTER)
        self.__label_widget = LabelPlayerWidget(True, 30)
        self.__label_widget.show()
        self.__label_widget.set_hexpand(True)
        self.__label_widget.set_vexpand(True)
        self.__label_widget.set_justify(Gtk.Justification.CENTER)
        eventbox = Gtk.EventBox.new()
        eventbox.show()
        eventbox.connect("button-release-event",
                         self.__on_image_button_release_event)
        eventbox.connect("realize", self.__on_image_realize)
        eventbox.add(self.__artwork_widget)
        self.__revealer = builder.get_object("revealer")
        self.__datetime = builder.get_object("datetime")
        self.__overlay_grid = builder.get_object("overlay_grid")
        widget = builder.get_object("widget")
        if App().settings.get_value("artist-artwork"):
            self.__overlay_grid.attach(self.__buttons_widget, 0, 4, 2, 1)
            self.__overlay_grid.attach(self.__label_widget, 0, 2, 2, 1)
            self.__overlay_grid.attach(self.__progress_widget, 0, 3, 2, 1)
            self.__overlay_grid.attach(eventbox, 2, 2, 1, 3)
            eventbox.set_margin_end(MARGIN_BIG)
            eventbox.set_property("valign", Gtk.Align.END)
            eventbox.set_property("halign", Gtk.Align.END)
            self.__artwork_widget.set_margin_end(MARGIN_BIG)
            self.__artwork_widget.set_margin_bottom(MARGIN_BIG)
            self.__label_widget.set_property("valign", Gtk.Align.END)
            self.__artwork_widget.set_property("valign", Gtk.Align.END)
        else:
            self.__overlay_grid.attach(self.__buttons_widget, 0, 4, 3, 1)
            self.__overlay_grid.attach(self.__label_widget, 0, 2, 3, 1)
            self.__overlay_grid.attach(self.__progress_widget, 0, 3, 3, 1)
            self.__overlay_grid.attach(eventbox, 0, 1, 3, 1)
            eventbox.set_vexpand(True)
        close_btn = builder.get_object("close_btn")
        preferences = Gio.Settings.new("org.gnome.desktop.wm.preferences")
        layout = preferences.get_value("button-layout").get_string()
        if layout.split(":")[0] == "close":
            self.__overlay_grid.attach(close_btn, 0, 0, 1, 1)
            close_btn.set_property("halign", Gtk.Align.START)
        else:
            self.__overlay_grid.attach(close_btn, 2, 0, 1, 1)
            close_btn.set_property("halign", Gtk.Align.END)
        self._artwork = builder.get_object("cover")
        self.connect("key-release-event", self.__on_key_release_event)
        # Add a navigation widget on the right
        self.__back_button = Gtk.Button.new_from_icon_name(
            "go-previous-symbolic", Gtk.IconSize.BUTTON)
        self.__back_button.set_sensitive(False)
        self.__back_button.set_relief(Gtk.ReliefStyle.NONE)
        self.__back_button.set_property("valign", Gtk.Align.START)
        self.__back_button.set_property("halign", Gtk.Align.START)
        self.__back_button.connect("clicked", self.__on_back_button_clicked)
        self.__back_button.set_margin_start(5)
        self.__back_button.set_margin_end(5)
        self.__back_button.set_margin_top(5)
        self.__back_button.set_margin_bottom(5)
        self.__back_button.show()
        self.__background_artwork = builder.get_object("background_artwork")
        self.__container = Container()
        self.__container.show()
        self.__container.setup()
        self.__sidebar = Gtk.Grid()
        self.__sidebar.set_size_request(400, -1)
        self.__sidebar.set_orientation(Gtk.Orientation.VERTICAL)
        self.__sidebar.get_style_context().add_class("borders-left-top")
        self.__sidebar.show()
        self.__sidebar.add(self.__back_button)
        self.__sidebar.add(self.__container)
        self.__sidebar.set_size_request(450, -1)
        self.__sidebar.get_style_context().add_class("background-opacity")
        self.__container.connect("can-go-back-changed",
                                 self.__on_can_go_back_changed)
        self.connect("size-allocate", self.__on_size_allocate)
        self.__revealer.add(self.__sidebar)
        self.add(widget)

    def do_show(self):
        """
            Setup window for current screen
        """
        Gtk.Window.do_show(self)
        self.__buttons_widget.update()
        self.__label_widget.update()
        self.__progress_widget.update()
        self.__update_progress_visibility()
        self.__setup_artwork_type()
        if self.__timeout_id is None:
            try:
                interface = Gio.Settings.new("org.gnome.desktop.interface")
                show_seconds = interface.get_value("clock-show-seconds")
            except:
                show_seconds = False
            self.__update_datetime(show_seconds)
            self.__timeout_id = GLib.timeout_add(1000,
                                                 self.__update_datetime,
                                                 show_seconds)
        self.__progress_widget.update_position(
            App().player.position / Gst.SECOND)
        screen = Gdk.Screen.get_default()
        monitor = screen.get_monitor_at_window(App().main_window.get_window())
        self.fullscreen_on_monitor(screen, monitor)
        # Disable screensaver (idle)
        App().inhibitor.manual_inhibit(
                Gtk.ApplicationInhibitFlags.IDLE |
                Gtk.ApplicationInhibitFlags.SUSPEND)

    def do_hide(self):
        """
            Clean window
        """
        Gtk.Window.do_hide(self)
        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)
            self.__timeout_id = None
        App().inhibitor.manual_uninhibit()

    @property
    def miniplayer(self):
        return App().main_window.miniplayer

    @property
    def folded(self):
        return False

    @property
    def toolbar(self):
        return App().main_window.toolbar

    @property
    def container(self):
        """
            Get container
            @return Container
        """
        return self.__container

#######################
# PROTECTED           #
#######################
    def _on_close_button_clicked(self, button):
        """
            Destroy self
            @param button as Gtk.Button
        """
        self.destroy()

    def _on_reveal_button_clicked(self, button):
        """
            Reveal widget
            @param button as Gtk.Button
        """
        if self.__revealer.get_reveal_child():
            self.__revealer.set_reveal_child(False)
            button.get_image().set_from_icon_name("pan-start-symbolic",
                                                  Gtk.IconSize.BUTTON)
        else:
            self.__revealer.set_reveal_child(True)
            button.get_image().set_from_icon_name("pan-end-symbolic",
                                                  Gtk.IconSize.BUTTON)

    def _on_current_changed(self, player):
        """
            Update background
            @param player as Player
        """
        self.__update_background()
        self.__update_progress_visibility()

#######################
# PRIVATE             #
#######################
    def __setup_artwork_type(self):
        """
            Setup artwork type
        """
        fs_type = App().settings.get_value("fullscreen-type").get_int32()
        context = self.__artwork_widget.get_style_context()
        behaviour = ArtBehaviour.CROP_SQUARE
        if fs_type & ArtBehaviour.ROUNDED:
            context.add_class("image-rotate")
            context.remove_class("small-cover-frame")
            behaviour |= ArtBehaviour.ROUNDED
        else:
            context.remove_class("image-rotate")
            context.add_class("small-cover-frame")
        self.__artwork_widget.set_behaviour(behaviour)
        self.__artwork_widget.update()

    def __update_progress_visibility(self):
        """
            Update progress bar visibility
        """
        if App().player.current_track.id is not None:
            self.__progress_widget.show()
        else:
            self.__progress_widget.hide()

    def __update_background(self, album_artwork=False):
        """
            Update window background
            @param album_artwork as bool
        """
        try:
            if App().player.current_track.id is None:
                return
            allocation = self.get_allocation()
            if allocation.width <= 1 or allocation.height <= 1:
                return
            behaviour = App().settings.get_value("fullscreen-type").get_int32()
            behaviour |= (ArtBehaviour.CROP |
                          ArtBehaviour.DARKER)
            # We don't want this for background, stored for album cover
            behaviour &= ~ArtBehaviour.ROUNDED
            if not album_artwork and\
                    App().settings.get_value("artist-artwork"):
                if App().player.current_track.album.artists:
                    artist = App().player.current_track.album.artists[0]
                else:
                    artist = App().player.current_track.artists[0]
                if self.__background_id == artist:
                    return
                App().art_helper.set_artist_artwork(
                                    artist,
                                    allocation.width,
                                    allocation.height,
                                    self.get_scale_factor(),
                                    behaviour,
                                    self.__on_artwork,
                                    False)
            else:
                if self.__background_id == App().player.current_track.album.id:
                    return
                App().art_helper.set_album_artwork(
                                    App().player.current_track.album,
                                    allocation.width,
                                    allocation.height,
                                    self.get_scale_factor(),
                                    ArtBehaviour.BLUR_MAX |
                                    ArtBehaviour.CROP |
                                    ArtBehaviour.DARKER,
                                    self.__on_artwork,
                                    True)
        except Exception as e:
            Logger.error("Fullscreen::__update_background(): %s", e)

    def __update_datetime(self, show_seconds=False):
        """
            Update datetime in headerbar
            @param show_seconds as bool
        """
        now = datetime.now()
        if show_seconds:
            self.__datetime.set_label(now.strftime("%a %d %b %X"))
        else:
            self.__datetime.set_label(now.strftime("%a %d %b %X")[:-3])
        if self.__timeout_id is None:
            self.__timeout_id = GLib.timeout_add(60000, self.__update_datetime)
            return False
        return True

    def __on_image_realize(self, eventbox):
        """
            Set cursor
            @param eventbox as Gtk.EventBox
        """
        try:
            eventbox.get_window().set_cursor(Gdk.Cursor(Gdk.CursorType.HAND2))
        except:
            Logger.warning(_("You are using a broken cursor theme!"))

    def __on_image_button_release_event(self, widget, event):
        """
            Change artwork type
            @param widget as Gtk.Widget
            @param event as Gdk.Event
        """
        fs_type = App().settings.get_value("fullscreen-type").get_int32()
        if fs_type & ArtBehaviour.BLUR_HARD and\
                fs_type & ArtBehaviour.ROUNDED:
            fs_type = ArtBehaviour.NONE
        elif fs_type & ArtBehaviour.NONE:
            fs_type = ArtBehaviour.BLUR_HARD
        elif fs_type & ArtBehaviour.BLUR_HARD:
            fs_type |= ArtBehaviour.ROUNDED
            fs_type &= ~ArtBehaviour.BLUR_HARD
        elif fs_type & ArtBehaviour.ROUNDED:
            fs_type |= ArtBehaviour.BLUR_HARD
        else:
            fs_type = ArtBehaviour.NONE
        App().settings.set_value("fullscreen-type",
                                 GLib.Variant("i", fs_type))
        self.__setup_artwork_type()
        self.__update_background()

    def __on_artwork(self, surface, album_artwork):
        """
            Set background artwork
            @param surface as str
            @param album_artwork as bool
        """
        if surface is None and not album_artwork:
            self.__update_background(True)
        else:
            self.__background_artwork.set_from_surface(surface)
            del surface

    def __on_key_release_event(self, widget, event):
        """
            Destroy window if Esc
            @param widget as Gtk.Widget
            @param event as Gdk.event
        """
        if event.keyval == Gdk.KEY_Escape:
            self.destroy()

    def __on_back_button_clicked(self, button):
        """
            Go back in container stack
            @param button as Gtk.Button
        """
        self.__container.go_back()

    def __on_can_go_back_changed(self, window, back):
        """
            Make button sensitive
            @param window as Gtk.Window
            @param back as bool
        """
        if back:
            self.__back_button.set_sensitive(True)
        else:
            self.__back_button.set_sensitive(False)

    def __on_size_allocate(self, widget, allocation):
        """
            Update background if needed
            @param widget as Gtk.Widget
            @param allocation as Gtk.Allocation
        """
        if allocation.width <= 1 or\
                allocation.height <= 1 or\
                allocation.width == self.__allocation.width or\
                allocation.height == self.__allocation.height:
            return
        self.__allocation = allocation
        size = allocation.width * ArtSize.FULLSCREEN / 1920
        self.__artwork_widget.set_art_size(size, size)
        self.__artwork_widget.update(True)
        self.__update_background()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, Gtk

from gettext import gettext as _

from lollypop.define import App, MARGIN_SMALL, Type


class PlaylistMenu(Gio.Menu):
    """
        A playlist menu
    """

    def __init__(self, playlist_id, view_type, header=False):
        """
            Init variable
            @param playlist_id as int
            @param view_type as ViewType
            @param header as bool
        """
        Gio.Menu.__init__(self)
        self.__playlist_id = playlist_id
        name = App().playlists.get_name(playlist_id)
        if header:
            from lollypop.menu_header import RoundedMenuHeader
            artwork_name = "playlist_%s" % name
            self.append_item(RoundedMenuHeader(name, artwork_name))
        menu = Gio.Menu()
        save_action = Gio.SimpleAction(name="save_pl_action")
        App().add_action(save_action)
        save_action.connect("activate", self.__on_save_action_activate)
        menu.append(_("Save playlist"), "app.save_pl_action")
        if App().ws_director.lastfm_ws is not None and\
                playlist_id == Type.LOVED:
            lastfm_action = Gio.SimpleAction(name="lastfm_action")
            App().add_action(lastfm_action)
            lastfm_action.connect("activate",
                                  self.__on_lastfm_action_activate)
            menu.append(_("Sync from Last.FM"), "app.lastfm_action")
        if playlist_id >= 0:
            if not App().playlists.get_track_uris(playlist_id):
                smart_action = Gio.SimpleAction(name="smart_action")
                App().add_action(smart_action)
                smart_action.connect("activate",
                                     self.__on_smart_action_activate)
                menu.append(_("Manage smart playlist"), "app.smart_action")
            remove_action = Gio.SimpleAction(name="remove_pl_action")
            App().add_action(remove_action)
            remove_action.connect("activate", self.__on_remove_action_activate)
            menu.append(_("Remove playlist"), "app.remove_pl_action")
        from lollypop.menu_playback import PlaylistPlaybackMenu
        playback_menu = PlaylistPlaybackMenu(playlist_id)
        self.append_section(_("Playlist"), playback_menu)
        section = Gio.Menu()
        self.append_section(_("Add to"), section)
        from lollypop.menu_sync import SyncPlaylistsMenu
        section.append_submenu(_("Devices"),
                               SyncPlaylistsMenu(playlist_id))
        self.append_section(_("Edit"), menu)

#######################
# PRIVATE             #
#######################
    def __on_smart_action_activate(self, action, variant):
        """
            Show smart playlist editor
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        App().window.container.show_smart_playlist_editor(self.__playlist_id)

    def __on_lastfm_action_activate(self, action, variant):
        """
            Sync playlist with Last.FM
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        App().ws_director.lastfm_ws.sync_loved_tracks()

    def __on_remove_action_activate(self, action, variant):
        """
            Remove playlist
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        def remove_playlist():
            App().playlists.remove(self.__playlist_id)
        from lollypop.app_notification import AppNotification
        notification = AppNotification(_("Remove this playlist?"),
                                       [_("Confirm")],
                                       [remove_playlist])
        notification.show()
        App().window.container.add_overlay(notification)
        notification.set_reveal_child(True)

    def __on_save_action_activate(self, action, variant):
        """
            Save playlist to file
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        filechooser = Gtk.FileChooserNative.new(_("Save playlist"),
                                                App().window,
                                                Gtk.FileChooserAction.SAVE,
                                                _("Save"),
                                                _("Cancel"))
        filter = Gtk.FileFilter.new()
        filter.set_name("audio/x-mpegurl")
        filter.add_mime_type("audio/x-mpegurl")
        filechooser.add_filter(filter)
        filechooser.set_do_overwrite_confirmation(True)
        name = App().playlists.get_name(self.__playlist_id)
        filechooser.set_current_name("%s.m3u" % name)
        filechooser.connect("response", self.__on_save_response)
        filechooser.run()

    def __on_save_response(self, dialog, response_id):
        """
            Save playlist
            @param dialog as Gtk.NativeDialog
            @param response_id as int
        """
        if response_id == Gtk.ResponseType.ACCEPT:
            uri = dialog.get_file().get_uri()
            App().playlists.set_sync_uri(self.__playlist_id, uri)
            App().playlists.sync_to_disk(self.__playlist_id, True)


class PlaylistMenuExt(Gtk.Grid):
    """
        Additional widgets for playlist menu
    """

    def __init__(self, playlist_id):
        """
            Init widget
            @param playlist_id as int
        """
        Gtk.Grid.__init__(self)
        self.set_margin_top(MARGIN_SMALL)
        self.set_row_spacing(MARGIN_SMALL)
        self.set_orientation(Gtk.Orientation.VERTICAL)

        entry = Gtk.Entry()
        entry.set_margin_top(MARGIN_SMALL)
        entry.set_margin_start(MARGIN_SMALL)
        entry.set_margin_end(MARGIN_SMALL)
        entry.set_margin_bottom(MARGIN_SMALL)
        entry.set_property("hexpand", True)
        entry.set_text(App().playlists.get_name(playlist_id))
        entry.connect("changed", self.__on_entry_changed, playlist_id)
        entry.show()
        self.add(entry)

    @property
    def section(self):
        return None

    @property
    def submenu_name(self):
        return None

#######################
# PRIVATE             #
#######################
    def __on_entry_changed(self, entry, playlist_id):
        """
            Update playlist name
            @param entry as Gtk.Entry
            @param playlist_id as int
        """
        new_name = entry.get_text()
        App().playlists.rename(playlist_id, new_name)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib


class SizeAllocationHelper:
    """
        Listen to size-allocate signal and ignore unwanted allocations
    """

    def __init__(self, use_parent=False):
        """
            Init helper
            @param use_parent as bool: follow parent size allocation
        """
        self.__allocation_timeout_id = None
        self.__width = self.__height = 0
        if use_parent:
            parent = self.get_parent()
            parent.connect("size-allocate", self.__on_size_allocate)
        else:
            self.connect("size-allocate", self.__on_size_allocate)

    @property
    def width(self):
        """
            Get widget width
            @return int
        """
        return self.__width

    @property
    def height(self):
        """
            Get widget height
            @return int
        """
        return self.__height

#######################
# PROTECTED           #
#######################
    def _handle_width_allocate(self, allocation):
        """
            @param allocation as Gtk.Allocation
            @return True if allocation is valid
        """
        if allocation.width == 1 or self.__width == allocation.width:
            return False
        self.__width = allocation.width
        return True

    def _handle_height_allocate(self, allocation):
        """
            @param allocation as Gtk.Allocation
            @return True if allocation is valid
        """
        if allocation.height == 1 or self.__height == allocation.height:
            return False
        self.__height = allocation.height
        return True

#######################
# PRIVATE             #
#######################
    def __handle_size_allocate(self, allocation):
        """
            Pass allocation to width/height handler
        """
        self.__allocation_timeout_id = None
        self._handle_width_allocate(allocation)
        self._handle_height_allocate(allocation)

    def __on_size_allocate(self, widget, allocation):
        """
            Filter unwanted allocations
            @param widget as Gtk.Widget
            @param allocation as Gtk.Allocation
        """
        if self.__allocation_timeout_id is not None:
            GLib.source_remove(self.__allocation_timeout_id)
        self.__allocation_timeout_id = GLib.idle_add(
            self.__handle_size_allocate, allocation)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Pango

from pickle import load, dump
from gettext import gettext as _

from lollypop.define import App, ArtSize, MARGIN, ViewType, StorageType
from lollypop.define import ArtBehaviour, LOLLYPOP_DATA_PATH
from lollypop.widgets_banner import BannerWidget
from lollypop.widgets_cover import BrowsableCoverWidget
from lollypop.objects_album import Album
from lollypop.logger import Logger
from lollypop.utils import get_default_storage_type, popup_widget
from lollypop.helper_signals import SignalsHelper, signals_map


class TodayBannerWidget(BannerWidget, SignalsHelper):
    """
        Banner for today album
    """

    def get_today_album():
        """
            Get today album
            @return Album/None
        """
        current_date = GLib.DateTime.new_now_local().get_day_of_year()
        (date, album_id) = (0, None)
        try:
            (date, album_id) = load(
                open(LOLLYPOP_DATA_PATH + "/today.bin", "rb"))
            if App().albums.get_storage_type(album_id) == StorageType.NONE:
                date = 0
        except Exception as e:
            Logger.warning("TodayBannerWidget::__get_today_album(): %s", e)
        try:
            if date != current_date:
                storage_type = get_default_storage_type()
                album_id = App().albums.get_randoms(
                    storage_type, None, False, 1)[0]
                dump((current_date, album_id),
                     open(LOLLYPOP_DATA_PATH + "/today.bin", "wb"))
            return Album(album_id)
        except Exception as e:
            Logger.error("TodayBannerWidget::__get_today_album(): %s", e)
        return None

    @signals_map
    def __init__(self, album, view_type):
        """
            Init cover widget
            @param album
            @param view_type as ViewType
        """
        BannerWidget.__init__(self, view_type | ViewType.OVERLAY)
        self.__album = album
        album_name = GLib.markup_escape_text(self.__album.name)
        self.__title_label = Gtk.Label.new()
        self.__title_label.show()
        markup = _("<b>Album of the day</b>\n")
        markup += "<span size='small' alpha='40000'>%s</span>\n" % album_name
        artist_name = GLib.markup_escape_text(", ".join(self.__album.artists))
        markup += "<span size='x-small' alpha='40000'>%s</span>" % artist_name
        self.__title_label.set_markup(markup)
        self.__title_label.set_ellipsize(Pango.EllipsizeMode.END)
        self.__title_label.set_xalign(0.0)
        self.__title_label.set_vexpand(True)
        self.__title_label.set_margin_start(MARGIN)
        self.__cover_widget = BrowsableCoverWidget(self.__album, view_type)
        self.__cover_widget.show()
        play_button = Gtk.Button.new()
        play_button.show()
        play_button.set_property("valign", Gtk.Align.CENTER)
        image = Gtk.Image.new()
        image.show()
        play_button.set_image(image)
        play_button.connect("clicked", self.__on_play_button_clicked)
        play_button.get_style_context().add_class("banner-button")
        play_button.set_property("halign", Gtk.Align.END)
        play_button.set_hexpand(True)
        play_button.get_image().set_from_icon_name(
            "media-playback-start-symbolic", Gtk.IconSize.BUTTON)
        menu_button = Gtk.Button.new_from_icon_name(
            "view-more-symbolic", Gtk.IconSize.BUTTON)
        menu_button.show()
        menu_button.set_property("valign", Gtk.Align.CENTER)
        menu_button.get_style_context().add_class("banner-button")
        menu_button.set_property("halign", Gtk.Align.END)
        menu_button.connect("clicked", self.__on_menu_button_clicked)
        grid = Gtk.Grid()
        grid.show()
        grid.set_margin_start(MARGIN)
        grid.set_margin_end(MARGIN)
        grid.add(self.__cover_widget)
        grid.add(self.__title_label)
        box = Gtk.Box()
        box.show()
        box.get_style_context().add_class("linked")
        box.add(play_button)
        box.add(menu_button)
        grid.add(box)
        self._overlay.add_overlay(grid)
        self._overlay.set_overlay_pass_through(grid, True)
        self.__set_internal_size()
        return [
                (App().window.container.widget, "notify::folded",
                 "_on_container_folded"),
                (App().album_art, "album-artwork-changed",
                 "_on_album_artwork_changed")
        ]

    def update_for_width(self, width):
        """
            Update banner internals for width, call this before showing banner
            @param width as int
        """
        BannerWidget.update_for_width(self, width)
        self.__set_artwork()

#######################
# PROTECTED           #
#######################
    def _handle_width_allocate(self, allocation):
        """
            Update artwork
            @param allocation as Gtk.Allocation
        """
        if BannerWidget._handle_width_allocate(self, allocation):
            self.__set_artwork()

    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        self.__set_internal_size()

    def _on_album_artwork_changed(self, art, album_id):
        """
            Update cover for album_id
            @param art as Art
            @param album_id as int
        """
        if album_id == self.__album.id:
            self.__set_artwork()

#######################
# PRIVATE             #
#######################
    def __set_artwork(self):
        """
            Set artwork on banner
        """
        if App().animations:
            App().art_helper.set_album_artwork(
                    self.__album,
                    # +100 to prevent resize lag
                    self.width + 100,
                    self.height,
                    self._artwork.get_scale_factor(),
                    ArtBehaviour.BLUR_HARD |
                    ArtBehaviour.DARKER,
                    self._on_artwork)

    def __set_internal_size(self):
        """
            Set content size based on current width
        """
        # Text size
        title_context = self.__title_label.get_style_context()
        for c in title_context.list_classes():
            title_context.remove_class(c)
        if App().window.folded:
            art_size = ArtSize.MEDIUM
            cls = "text-large"
        else:
            art_size = ArtSize.BANNER
            cls = "text-x-large"
        self.__title_label.get_style_context().add_class(cls)
        self.__cover_widget.set_art_size(art_size)

    def __on_play_button_clicked(self, button):
        """
            Play album
            @param button as Gtk.Button
        """
        App().player.play_album(self.__album)

    def __on_menu_button_clicked(self, button):
        """
            Show suggestions menu
            @param button as Gtk.Button
        """
        from lollypop.menu_suggestions import SuggestionsMenu
        from lollypop.widgets_menu import MenuBuilder
        menu = SuggestionsMenu(App().window.folded)
        menu_widget = MenuBuilder(menu)
        menu_widget.show()
        popup_widget(menu_widget, button, None, None, button)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Gio, Pango

from gettext import gettext as _
import re

from lollypop.define import App, ViewType, MARGIN
from lollypop.define import ARTISTS_PATH
from lollypop.objects_album import Album
from lollypop.information_store import InformationStore
from lollypop.view_albums_list import AlbumsListView
from lollypop.view import View
from lollypop.utils import get_default_storage_type
from lollypop.widgets_banner_information import InformationBannerWidget
from lollypop.helper_signals import SignalsHelper, signals_map


class ArtistRow(Gtk.ListBoxRow):
    """
        Artist row for Wikipedia
    """

    def __init__(self, item):
        """
            Init row
            @param item as (str, str, str)
        """
        Gtk.ListBoxRow.__init__(self)
        self.__locale = item[0]
        self.__page_id = item[2]
        label = Gtk.Label.new("%s: %s" % (item[0], item[1]))
        label.set_property("halign", Gtk.Align.START)
        label.set_ellipsize(Pango.EllipsizeMode.END)
        label.get_style_context().add_class("padding")
        label.show()
        self.add(label)

    @property
    def locale(self):
        """
            Get locale
            @return str
        """
        return self.__locale

    @property
    def page_id(self):
        """
            Get page id
            @return str
        """
        return self.__page_id


class InformationView(View, SignalsHelper):
    """
        View with artist information
    """

    @signals_map
    def __init__(self, view_type, minimal=False):
        """
            Init artist infos
            @param view_type as ViewType
            @param minimal as bool
        """
        View.__init__(self, get_default_storage_type(),
                      view_type | ViewType.OVERLAY)
        self.__information_store = InformationStore()
        self.__cancellable = Gio.Cancellable()
        self.__minimal = minimal
        self.__artist_name = ""
        if minimal:
            return []
        else:
            return [
                (App().window, "notify::folded", "_on_container_folded"),
            ]

    def populate(self, artist_id=None):
        """
            Show information for artists
            @param artist_id as int
        """
        self.__banner = None
        artist_ids = App().player.current_track.artist_ids\
            if artist_id is None else [artist_id]
        self.__artist_name = App().artists.get_name(artist_ids[0])
        if not self.__minimal:
            storage_type = App().player.current_track.storage_type
            self.__banner = InformationBannerWidget(artist_ids[0])
            self.__banner.show()
            self.__banner.connect("search", self.__on_banner_search)
            self.__albums_view = AlbumsListView([], [],
                                                ViewType.SCROLLED |
                                                ViewType.ARTIST)
            self.__albums_view.set_size_request(350, -1)
            self.__albums_view.show()
            self.__albums_view.set_halign(Gtk.Align.END)
            self.__albums_view.add_widget(self.__albums_view.box)
            albums = []
            storage_type = get_default_storage_type()
            for album_id in App().albums.get_ids([], artist_ids,
                                                 storage_type, True):
                albums.append(Album(album_id))
            if not albums:
                albums = [App().player.current_track.album]
            self.__albums_view.populate(albums)
        self.__listbox = Gtk.ListBox.new()
        self.__listbox.show()
        self.__listbox.get_style_context().add_class("trackswidget")
        self.__listbox.connect("row-activated", self.__on_row_activated)
        self.__listbox.set_hexpand(True)
        self.__listbox.set_halign(Gtk.Align.CENTER)
        self.__label = Gtk.Label.new()
        self.__label.show()
        self.__label.set_justify(Gtk.Justification.FILL)
        self.__label.set_line_wrap(True)
        self.__label.set_line_wrap_mode(Pango.WrapMode.WORD)
        self.__label.set_valign(Gtk.Align.START)
        self.__label.set_margin_top(MARGIN)
        self.__label.set_margin_bottom(MARGIN)
        self.__label.set_margin_start(MARGIN)
        self.__label.set_margin_end(MARGIN)
        self.__label.set_hexpand(True)
        widget = Gtk.Grid()
        widget.show()
        self.__stack = Gtk.Stack.new()
        self.__stack.show()
        self.__stack.add(self.__label)
        self.__stack.add(self.__listbox)
        self.add_widget(widget, self.__banner)
        widget.add(self.__stack)
        if not self.__minimal:
            widget.add(self.__albums_view)
        self._on_container_folded()
        content = self.__information_store.get_information(self.__artist_name,
                                                           ARTISTS_PATH)
        if content is None:
            self.__label.set_text(_("Loading information"))
            from lollypop.information_downloader import InformationDownloader
            downloader = InformationDownloader()
            downloader.get_information(self.__artist_name,
                                       self.__on_artist_information,
                                       self.__artist_name)
        else:
            App().task_helper.run(self.__to_markup, content,
                                  callback=(self.__label.set_markup,))

    @property
    def artist_name(self):
        """
            Get view artist
            @return artist_name as str
        """
        return self.__artist_name

#######################
# PROTECTED           #
#######################
    def _on_unmap(self, widget):
        """
            Cancel operations
            @param widget as Gtk.Widget
        """
        self.__cancellable.cancel()

    def _on_container_folded(self, *ignore):
        """
            Handle libhandy folded status
        """
        if not self.__minimal:
            if App().window.folded:
                if not self.view_type & ViewType.SMALL:
                    self.__label.set_margin_start(MARGIN)
                    self.__label.set_margin_end(MARGIN)
                    self.__label.get_style_context().remove_class("text-large")
                self.__albums_view.hide()
            else:
                if not self.view_type & ViewType.SMALL:
                    self.__label.set_margin_start(100)
                    self.__label.set_margin_end(100)
                    self.__label.get_style_context().add_class("text-large")
                self.__albums_view.show()

#######################
# PRIVATE             #
#######################
    def __show_main_widget(self):
        """
            Show main widget in stack
        """
        self.__listbox.hide()
        self.__label.show()
        self.__stack.set_visible_child(self.__label)

    def __to_markup(self, data):
        """
            Transform message to markup
            @param data as bytes
            @return str
        """
        pango = ["large", "x-large", "xx-large"]
        start = ["^===*", "^==", "^="]
        end = ["===*$", "==$", "=$"]
        i = 0
        text = GLib.markup_escape_text(data.decode("utf-8"))
        while i < len(pango):
            text = re.sub(start[i], "<b><span size='%s'>" % pango[i],
                          text, flags=re.M)
            text = re.sub(end[i], "</span></b>", text, flags=re.M)
            i += 1
        return text

    def __on_banner_search(self, banner, status):
        """
            Search for artist
            @param banner as InformationBannerwidget
            @param status as bool
        """
        if status:
            from lollypop.helper_web_wikipedia import WikipediaHelper
            wikipedia = WikipediaHelper()
            self.__label.hide()
            self.__listbox.show()
            self.__stack.set_visible_child(self.__listbox)
            App().task_helper.run(
                wikipedia.get_search_list,
                self.__artist_name,
                callback=(self.__on_wikipedia_search_list,))
        else:
            self.__show_main_widget()

    def __on_wikipedia_search_list(self, items):
        """
            Populate view with items
            @param items as [(str, str)]
        """
        if items:
            for item in items:
                row = ArtistRow(item)
                row.show()
                self.__listbox.add(row)
        else:
            self.__show_main_widget()
            self.__label.set_text(
                _("No information for %s") % self.__artist_name)

    def __on_wikipedia_get_content(self, content):
        """
            Update label and save to cache
            @param content as str
        """
        if content is not None:
            App().task_helper.run(self.__to_markup, content,
                                  callback=(self.__label.set_markup,))
            self.__information_store.save_information(
                self.__artist_name, ARTISTS_PATH, content)

    def __on_artist_artwork(self, surface):
        """
            Finish widget initialisation
            @param surface as cairo.Surface
        """
        if surface is None:
            self.__artist_artwork.hide()
        else:
            self.__artist_artwork.set_from_surface(surface)

    def __on_artist_information(self, content, artist_name):
        """
            Set label
            @param content as bytes
            @param artist_name as str
        """
        if content is None:
            self.__show_main_widget()
            self.__label.set_text(
                _("No information for %s") % self.__artist_name)
        else:
            App().task_helper.run(self.__to_markup, content,
                                  callback=(self.__label.set_markup,))
            self.__information_store.save_information(self.__artist_name,
                                                      ARTISTS_PATH,
                                                      content)

    def __on_row_activated(self, listbox, row):
        """
            Update artist information
            @param listbox as Gtk.ListBox
            @param row as Gtk.ListBoxRow
        """
        self.__show_main_widget()
        self.__banner.button.set_active(False)
        from lollypop.helper_web_wikipedia import WikipediaHelper
        wikipedia = WikipediaHelper()
        App().task_helper.run(wikipedia.get_content_for_page_id,
                              row.page_id, row.locale,
                              callback=(self.__on_wikipedia_get_content,))



from locale import strcoll

_LETTERS = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'J', 'K', 'L', 'M',
            'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'W', 'X', 'Y', 'Z', 'Z']
_CHARS = ['阿', '八', '嚓', '哒', '妸', '发', '旮', '哈', '讥', '咔', '垃', '痳',
          '拏', '噢', '妑', '七', '呥', '扨', '它', '穵', '夕', '丫', '帀', '做']
_RANGES = [(l, c, cc) for ((l, c), (ll, cc)) in
           zip(list(zip(_LETTERS, _CHARS))[:-1], list(zip(_LETTERS, _CHARS))[1:])]
_EXCEPTIONS = {'曾': 'Z', '沈': 'S', '酢': 'Z'}

def index_of(string):
    """
        Return the (Pinyin-ized) index character in upper case.
        Return '#' if it's a digit.
        Return '…' if it's not recognizable in this locale.
        Locale: zh_CN.
        @param string as str
        @return str
    """
    head = string[0]

    if head.upper() in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
        return head.upper()

    if head in '1234567890':
        return '#'

    if head in _EXCEPTIONS:
        return _EXCEPTIONS[head]

    for letter, char, next_char in _RANGES:
        if strcoll(char, head) <= 0 and strcoll(head, next_char) <= 0:
            return letter

    return '?'



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gst, GLib, GstAudio

from time import sleep

from lollypop.define import App


class TransitionsPlayer:
    """
        Handle track transitions
    """
    __PADDING = 250

    def __init__(self):
        """
            Init playbin
        """
        self.__crossfading_id = None
        self.__crossfade_up = False
        self.__crossfade_down = False
        self.update_crossfading()

    def load(self, track):
        """
            Load track and play it with transitions
            @param track as Track
            @return bool: True if track has been loaded
        """
        if self.crossfading and\
           self._current_track.id is not None and\
           self.is_playing:
            transition_duration = App().settings.get_value(
                    "transitions-duration").get_int32()
            self.__do_crossfade(transition_duration, track)
            return True
        return False

    def set_crossfading(self, status):
        """
            Set crossfading on/off
            @param status as bool
        """
        if status and self.__crossfading_id is None:
            transition_duration = App().settings.get_value(
                "transitions-duration").get_int32()
            timemout = min(transition_duration, 500)
            self.__crossfading_id = GLib.timeout_add(
                timemout, self.__check_for_crossfading)
        elif not status and self.__crossfading_id is not None:
            GLib.source_remove(self.__crossfading_id)
            self.__crossfading_id = None

    def update_crossfading(self):
        """
            Calculate if crossfading is needed
        """
        transitions = App().settings.get_value("transitions")
        party_only = App().settings.get_value("transitions-party-only")
        self.set_crossfading((transitions and not party_only) or
                             (transitions and party_only and self.is_party))

    @property
    def crossfading(self):
        """
            True if crossfading is on
            @return bool
        """
        return self.__crossfading_id is not None

#######################
# PRIVATE             #
#######################
    def __check_for_crossfading(self):
        """
            Check if we need to do crossfading
        """
        if self._current_track.duration > 0:
            remaining = self.remaining
            transition_duration = App().settings.get_value(
                    "transitions-duration").get_int32()
            if remaining < transition_duration + self.__PADDING:
                self.__do_crossfade(transition_duration,
                                    self._next_track)
        return True

    def __volume_up(self, playbin, plugins, duration):
        """
            Make volume going up smoothly
            @param playbin as Gst.Bin
            @param plugins as PluginsPlayer
            @param duration as int
        """
        plugins.volume.props.volume = 0.0
        self.__crossfade_up = True
        # We add padding because user will not hear track around 0.2
        sleep_ms = (duration + self.__PADDING) / 10
        while plugins.volume.props.volume < 1.0:
            vol = round(plugins.volume.props.volume + 0.1, 1)
            plugins.volume.props.volume = vol
            sleep(sleep_ms / 1000)
        self.__crossfade_up = False

    def __volume_down(self, playbin, plugins, duration):
        """
            Make volume going down smoothly
            @param playbin as Gst.Bin
            @param plugins as PluginsPlayer
            @param duration as int
        """
        plugins.volume.props.volume = 1.0
        self.__crossfade_down = True
        # We add padding because user will not hear track around 0.2
        sleep_ms = (duration + self.__PADDING) / 10
        while plugins.volume.props.volume > 0:
            vol = round(plugins.volume.props.volume - 0.1, 1)
            plugins.volume.props.volume = vol
            sleep(sleep_ms / 1000)
        playbin.set_state(Gst.State.NULL)
        self.__crossfade_down = False

    def __do_crossfade(self, duration, track):
        """
            Crossfade tracks
            @param duration as int
            @param track as Track
        """
        self._on_track_finished(self._current_track)

        if track.id is None:
            return

        # If some crossfade already running, just switch to track
        if self.__crossfade_up or self.__crossfade_down:
            self._playbin.set_state(Gst.State.NULL)
            if self._load_track(track):
                self.play()
            return

        App().task_helper.run(self.__volume_down, self._playbin,
                              self._plugins, duration)
        if self._playbin == self._playbin2:
            self._playbin = self._playbin1
            self._plugins = self._plugins1
        else:
            self._playbin = self._playbin2
            self._plugins = self._plugins2
        rate = App().settings.get_value("volume-rate").get_double()
        self._playbin.set_volume(GstAudio.StreamVolumeFormat.CUBIC, rate)
        if track.id is not None:
            self._playbin.set_state(Gst.State.NULL)
            if self._load_track(track):
                self._playbin.set_state(Gst.State.PLAYING)
            App().task_helper.run(self.__volume_up, self._playbin,
                                  self._plugins, duration)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

from lollypop.logger import Logger


class Settings(Gio.Settings):
    """
        Lollypop settings
    """

    def __init__(self):
        """
            Init settings
        """
        Gio.Settings.__init__(self)

    def new():
        """
            Return a new Settings object
        """
        settings = Gio.Settings.new("org.gnome.Lollypop")
        settings.__class__ = Settings
        return settings

    def get_music_uris(self):
        """
            Return music uris
            @return [str]
        """
        uris = self.get_value("music-uris")
        if not uris:
            filename = GLib.get_user_special_dir(
                GLib.UserDirectory.DIRECTORY_MUSIC)
            if filename:
                uris = [GLib.filename_to_uri(filename)]
            else:
                Logger.info("You need to add a music uri"
                            " to org.gnome.Lollypop in dconf")
        return list(uris)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio

from gettext import gettext as _

from lollypop.define import App, ViewType
from lollypop.utils import get_default_storage_type
from lollypop.objects_album import Album


class GenreMenu(Gio.Menu):
    """
        Contextual menu for a genres
    """
    def __init__(self, genre_id, view_type, header=False):
        """
            Init decade menu
            @param genre_id as int
            @param view_type as ViewType
            @param header as bool
        """
        Gio.Menu.__init__(self)
        if header:
            from lollypop.menu_header import RoundedMenuHeader
            name = App().genres.get_name(genre_id)
            artwork_name = "genre_%s" % name
            self.append_item(RoundedMenuHeader(name, artwork_name))
        if not view_type & ViewType.BANNER:
            from lollypop.menu_playback import GenrePlaybackMenu
            self.append_section(_("Playback"), GenrePlaybackMenu(genre_id))
        from lollypop.menu_sync import SyncAlbumsMenu
        section = Gio.Menu()
        self.append_section(_("Add to"), section)
        storage_type = get_default_storage_type()
        album_ids = App().albums.get_ids([genre_id], [],
                                         storage_type,
                                         False)
        album_ids += App().albums.get_compilation_ids([genre_id],
                                                      storage_type,
                                                      False)
        albums = [Album(album_id) for album_id in album_ids]
        section.append_submenu(_("Devices"), SyncAlbumsMenu(albums))



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib
from gi.repository.Gio import FILE_ATTRIBUTE_STANDARD_TYPE,\
    FILE_ATTRIBUTE_STANDARD_FAST_CONTENT_TYPE

from lollypop.define import App, ScanType
from lollypop.utils_file import is_audio
from lollypop.logger import Logger


SCAN_QUERY_INFO = "{},{}".format(FILE_ATTRIBUTE_STANDARD_TYPE,
                                 FILE_ATTRIBUTE_STANDARD_FAST_CONTENT_TYPE)


class Inotify:
    """
        Inotify support
    """
    # 2 seconds before updating database
    __TIMEOUT = 2000

    def __init__(self):
        """
            Init inode notification
        """
        self.__monitors = {}
        self.__collection_timeout_id = None
        self.__disable_timeout_id = None

    def add_monitor(self, uri):
        """
            Add a monitor for uri
            @param uri as string
        """
        # Check if there is already a monitor for this uri
        if uri in self.__monitors.keys():
            return
        try:
            f = Gio.File.new_for_uri(uri)
            monitor = f.monitor_directory(Gio.FileMonitorFlags.NONE,
                                          None)
            if monitor is not None:
                monitor.connect("changed", self.__on_dir_changed)
                self.__monitors[uri] = monitor
        except Exception as e:
            Logger.error("Inotify::add_monitor(): %s" % e)

    def disable(self, timeout=10000):
        """
            Disable inotify for timeout
            @param timeout as int
        """
        def on_timeout():
            self.__disable_timeout_id = None
        if self.__collection_timeout_id is not None:
            GLib.source_remove(self.__collection_timeout_id)
        if self.__disable_timeout_id is not None:
            GLib.source_remove(self.__disable_timeout_id)
        self.__disable_timeout_id = GLib.timeout_add(timeout, on_timeout)

#######################
# PRIVATE             #
#######################
    def __on_dir_changed(self, monitor, changed_file, other_file, event):
        """
            Stop collection scanner if running
            Delayed update by default
            @param monitor as Gio.FileMonitor
            @param changed_file as Gio.File/None
            @param other_file as Gio.File/None
            @param event as Gio.FileMonitorEvent
        """
        if self.__disable_timeout_id is not None:
            return
        try:
            changed_uri = changed_file.get_uri()
            # Do not monitor our self
            if changed_uri in self.__monitors.keys() and\
                    self.__monitors[changed_uri] == monitor:
                return

            if changed_file.query_exists():
                # Ignore non audio
                info = changed_file.query_info(SCAN_QUERY_INFO,
                                               Gio.FileQueryInfoFlags.NONE)
                if info.get_file_type() != Gio.FileType.DIRECTORY and\
                        is_audio(info):
                    return

            # Stop collection scanner and wait
            if App().scanner.is_locked():
                App().scanner.stop()
                GLib.timeout_add(self.__TIMEOUT,
                                 self.__on_dir_changed,
                                 monitor,
                                 changed_file,
                                 other_file,
                                 event)
            # Run update delayed
            else:
                if self.__collection_timeout_id is not None:
                    GLib.source_remove(self.__collection_timeout_id)
                if changed_file.has_parent():
                    uris = [changed_file.get_parent().get_uri()]
                else:
                    uris = [changed_uri]
                self.__collection_timeout_id = GLib.timeout_add(
                                                 self.__TIMEOUT,
                                                 self.__run_collection_update,
                                                 uris)
        except Exception as e:
            Logger.error("Inotify::__on_dir_changed()", e)

    def __run_collection_update(self, uris=[]):
        """
            Run a collection update
            @param uris as [str]
        """
        self.__collection_timeout_id = None
        App().scanner.update(ScanType.NEW_FILES, uris)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from lollypop.define import App, ViewType
from lollypop.view import View
from lollypop.utils import get_default_storage_type
from lollypop.helper_signals import SignalsHelper, signals_map


class InformationViewStack(View, SignalsHelper):
    """
        InformationView stack showing current track
    """

    @signals_map
    def __init__(self):
        """
            Init artist infos
        """
        View.__init__(self, get_default_storage_type(), ViewType.DEFAULT)
        self.__stack = Gtk.Stack.new()
        self.__stack.show()
        self.__stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.__stack.set_transition_duration(500)
        self.add_widget(self.__stack)
        self.__populate()
        return [
            (App().player, "current-changed", "_on_current_changed"),
        ]

#######################
# PROTECTED           #
#######################
    def _on_current_changed(self, player):
        """
            Update artist information
            @param player as Player
        """
        self.__populate()

#######################
# PRIVATE             #
#######################
    def __populate(self):
        """
            Add a new information view and destroy previous
        """
        if App().player.current_track.id is None:
            self.show_placeholder(True)
        else:
            self.show_placeholder(False)
            from lollypop.view_information import InformationView
            current_child = self.__stack.get_visible_child()
            if current_child is None or current_child.artist_name !=\
                    App().player.current_track.artists[0]:
                self.__view = InformationView(ViewType.SCROLLED, False)
                self.__view.show()
                self.__view.populate()
                self.__stack.add(self.__view)
                self.__stack.set_visible_child(self.__view)
                if current_child is not None:
                    GLib.timeout_add(500, current_child.destroy)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from lollypop.define import App
from lollypop.widgets_player_buttons import ButtonsPlayerWidget


class ToolbarPlayback(Gtk.Box):
    """
        Playback toolbar
    """

    def __init__(self, window):
        """
            Init toolbar
            @parma window as Window
        """
        Gtk.Box.__init__(self)
        self.__back_button = Gtk.Button.new_from_icon_name(
            "go-previous-symbolic", Gtk.IconSize.BUTTON)
        self.__back_button.show()
        self.__back_button.connect("clicked", self.__on_back_button_clicked)
        self.__player_buttons = ButtonsPlayerWidget()
        self.__player_buttons.show()
        self.set_spacing(10)
        self.pack_start(self.__back_button, False, False, 0)
        self.pack_start(self.__player_buttons, False, False, 0)
        window.container.connect("can-go-back-changed",
                                 self.__on_can_go_back_changed)

    @property
    def player_buttons(self):
        """
            Get player buttons
            @return Gtk.Box
        """
        return self.__player_buttons

    @property
    def back_button(self):
        """
            Get back button
            @return Gtk.Button
        """
        return self.__back_button

#######################
# PRIVATE             #
#######################
    def __on_back_button_clicked(self, button):
        """
            Go back in container stack
            @param button as Gtk.Button
        """
        App().window.container.go_back()

    def __on_can_go_back_changed(self, container, back):
        """
            Make button sensitive
            @param container as Container
            @param back as bool
        """
        if back:
            self.__back_button.set_sensitive(True)
        else:
            self.__back_button.set_sensitive(False)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Handy

from lollypop.define import App, ScanType, ArtSize
from lollypop.container import Container
from lollypop.toolbar import Toolbar
from lollypop.utils import emit_signal
from lollypop.helper_signals import SignalsHelper, signals_map
from lollypop.logger import Logger


class Window(Handy.ApplicationWindow, SignalsHelper):
    """
        Main window
    """

    @signals_map
    def __init__(self):
        """
            Init window
        """
        Handy.ApplicationWindow.__init__(self,
                                         application=App(),
                                         title="Lollypop",
                                         icon_name="org.gnome.Lollypop")
        self.__miniplayer = None
        self.__configure_timeout_id = None
        self.set_auto_startup_notification(False)
        self.connect("realize", self.__on_realize)
        # Does not work with a Gtk.Gesture in GTK3
        self.connect("button-release-event", self.__on_button_release_event)
        self.connect("window-state-event", self.__on_window_state_event)
        self.connect("configure-event", self.__on_configure_event)
        self.connect("destroy", self.__on_destroy)
        self.set_size_request(ArtSize.MINIPLAYER, ArtSize.MINIPLAYER)
        return [
            (App().player, "current-changed", "_on_current_changed")
        ]

    def setup(self):
        """
            Setup window content
        """
        self.__vgrid = Gtk.Grid()
        self.__vgrid.set_orientation(Gtk.Orientation.VERTICAL)
        self.__vgrid.show()
        self.__container = Container()
        self.__container.setup()
        self.__container.show()
        self.__toolbar = Toolbar(self)
        self.__toolbar.show()
        self.__vgrid.add(self.__toolbar)
        self.__toolbar.set_show_close_button(True)
        self.__vgrid.add(self.__container)
        self.add(self.__vgrid)
        self.__container.widget.connect("notify::folded",
                                        self.__on_container_folded)

    def show_miniplayer(self, show, reveal=False):
        """
            Show/hide subtoolbar
            @param show as bool
            @param reveal as bool
        """
        def show_buttons(show):
            if show:
                self.toolbar.end.show()
                self.toolbar.playback.show()
            else:
                self.toolbar.end.hide()
                self.toolbar.playback.hide()

        def on_revealed(miniplayer, revealed):
            miniplayer.set_vexpand(revealed)
            show_buttons(not revealed)
            if revealed:
                self.__toolbar.hide()
                self.__container.hide()
                emit_signal(self.__container, "can-go-back-changed", False)
            else:
                self.__toolbar.show()
                self.__container.show()
                emit_signal(self.__container, "can-go-back-changed",
                            self.__container.can_go_back)

        if show and self.__miniplayer is None:
            from lollypop.miniplayer import MiniPlayer
            self.__miniplayer = MiniPlayer()
            if App().player.current_track.id is not None:
                self.__miniplayer.show()
            self.__miniplayer.connect("revealed", on_revealed)
            self.__vgrid.add(self.__miniplayer)
            self.__miniplayer.set_vexpand(False)
        elif not show and self.__miniplayer is not None:
            if App().lookup_action("miniplayer").get_state():
                App().lookup_action("miniplayer").change_state(
                    GLib.Variant("b", False))
            else:
                if not self.folded:
                    self.__miniplayer.destroy()
                    self.__miniplayer = None
                self.__container.show()
                show_buttons(True)
        if self.__miniplayer is not None:
            if reveal:
                self.__miniplayer.reveal(True)
            else:
                self.__miniplayer.update_artwork()

    @property
    def folded(self):
        """
            True if window is adaptive, ie widget folded
        """
        return App().window.container.widget.get_folded()

    @property
    def miniplayer(self):
        """
            @return MiniPlayer
        """
        return self.__miniplayer

    @property
    def toolbar(self):
        """
            @return Toolbar
        """
        return self.__toolbar

    @property
    def container(self):
        """
            @return Container
        """
        return self.__container

##############
# PROTECTED  #
##############
    def _on_current_changed(self, player):
        """
            Update toolbar
            @param player as Player
        """
        if App().player.current_track.id is None:
            self.set_title("Lollypop")
        else:
            artists = ", ".join(player.current_track.artists)
            self.set_title("%s - %s" % (artists, player.current_track.name))
            if self.__miniplayer is not None:
                self.__miniplayer.show()

    def _on_configure_event_timeout(self, width, height, x, y):
        """
            Setup content based on current size
            @param width as int
            @param height as int
            @param x as int
            @param y as int
        """
        self.__configure_timeout_id = None
        if App().lookup_action("miniplayer").get_state():
            return
        if not self.is_maximized():
            App().settings.set_value("window-size",
                                     GLib.Variant("ai", [width, height]))
        App().settings.set_value("window-position", GLib.Variant("ai", [x, y]))

############
# PRIVATE  #
############
    def __setup_size_and_position(self):
        """
            Setup window position and size, callbacks
        """
        try:
            size = App().settings.get_value("window-size")
            pos = App().settings.get_value("window-position")
            self.resize(size[0], size[1])
            self.move(pos[0], pos[1])
            if App().settings.get_value("window-maximized"):
                self.maximize()
        except Exception as e:
            Logger.error("Window::__setup_size_and_position(): %s", e)

    def __on_realize(self, window):
        """
            Init window content
            @param window as Gtk.Window
        """
        self.__setup_size_and_position()
        if App().settings.get_value("auto-update") or App().tracks.is_empty():
            App().scanner.update(ScanType.FULL)
        # FIXME: remove this later
        elif GLib.file_test("/app", GLib.FileTest.EXISTS) and\
                not App().settings.get_value("flatpak-access-migration"):
            App().scanner.update(ScanType.FULL)

    def __on_button_release_event(self, window, event):
        """
            Handle special mouse buttons
            @param window as Gtk.Window
            @param event as Gdk.EventButton
        """
        if event.button == 8:
            App().window.container.go_back()
            return True

    def __on_window_state_event(self, widget, event):
        """
            Save maximised state
        """
        if not App().lookup_action("miniplayer").get_state():
            App().settings.set_boolean("window-maximized",
                                       "GDK_WINDOW_STATE_MAXIMIZED" in
                                       event.new_window_state.value_names)

    def __on_container_folded(self, leaflet, folded):
        """
            show/hide miniplayer
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        emit_signal(self.__container,
                    "can-go-back-changed",
                    self.__container.can_go_back)
        self.show_miniplayer(App().window.folded)

    def __on_destroy(self, widget):
        """
            Remove ref cycle, just to prevent output on DEBUG_LEAK
            @param widget as Gtk.Widget
        """
        self.__toolbar = None

    def __on_configure_event(self, window, event):
        """
            Delay event
            @param window as Gtk.Window
            @param event as Gdk.EventConfigure
        """
        if self.__configure_timeout_id:
            GLib.source_remove(self.__configure_timeout_id)
        (width, height) = window.get_size()
        (x, y) = window.get_position()
        self.__configure_timeout_id = GLib.idle_add(
            self._on_configure_event_timeout,
            width, height, x, y, priority=GLib.PRIORITY_LOW)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gio, GLib

from gettext import gettext as _
from random import choice

from lollypop.utils import set_cursor_type, on_query_tooltip, popup_widget
from lollypop.utils_artist import add_artist_to_playback, play_artists
from lollypop.define import App, ArtSize, ArtBehaviour, ViewType
from lollypop.widgets_banner import BannerWidget
from lollypop.helper_signals import SignalsHelper, signals_map


class ArtistBannerWidget(BannerWidget, SignalsHelper):
    """
        Banner for artist
    """

    @signals_map
    def __init__(self, genre_ids, artist_ids, storage_type, view_type):
        """
            Init artist banner
            @parma genre_ids as [int]
            @param artist_ids as [int]
            @param storage_type as StorageType
            @param view_type as ViewType (Unused)
        """
        BannerWidget.__init__(self, view_type | ViewType.OVERLAY)
        self.__genre_ids = genre_ids
        self.__artist_ids = artist_ids
        self.__storage_type = storage_type
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/ArtistBannerWidget.ui")
        builder.connect_signals(self)
        self.__badge_artwork = builder.get_object("badge_artwork")
        self.__labels = builder.get_object("label_event")
        self.__title_label = builder.get_object("artist")
        self.__title_label.connect("realize", set_cursor_type)
        self.__title_label.connect("query-tooltip", on_query_tooltip)
        self.__title_label.set_property("has-tooltip", True)
        self.__add_button = builder.get_object("add_button")
        self.__play_button = builder.get_object("play_button")
        self.__menu_button = builder.get_object("menu_button")
        if len(artist_ids) > 1:
            self.__menu_button.hide()
        builder.get_object("artwork_event").connect(
            "realize", set_cursor_type)
        self.__labels.connect("realize", set_cursor_type)
        self.__widget = builder.get_object("widget")
        artists = []
        for artist_id in self.__artist_ids:
            artists.append(App().artists.get_name(artist_id))
        self.__title_label.set_markup(GLib.markup_escape_text(
            ", ".join(artists)))
        self.__show_artwork = len(artist_ids) == 1
        self.__title_label.get_style_context().add_class("text-x-large")
        self._overlay.add_overlay(self.__widget)
        self._overlay.set_overlay_pass_through(self.__widget, True)
        self.__update_add_button()
        self.__set_internal_size()
        return [
               (App().window.container.widget, "notify::folded",
                "_on_container_folded"),
               (App().artist_art, "artist-artwork-changed",
                "_on_artist_artwork_changed"),
               (App().player, "playback-added", "_on_playback_changed"),
               (App().player, "playback-updated", "_on_playback_changed"),
               (App().player, "playback-setted", "_on_playback_changed"),
               (App().player, "playback-removed", "_on_playback_changed"),
               (App().settings, "changed::artist-artwork",
                "_on_artist_artwork_setting_changed")

        ]

    def update_for_width(self, width):
        """
            Update banner internals for width, call this before showing banner
            @param width as int
        """
        BannerWidget.update_for_width(self, width)
        self.__set_artwork()

#######################
# PROTECTED           #
#######################
    def _handle_width_allocate(self, allocation):
        """
            Update artwork
            @param allocation as Gtk.Allocation
        """
        if BannerWidget._handle_width_allocate(self, allocation):
            self.__set_artwork()

    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        self.__set_internal_size()

    def _on_artist_artwork_setting_changed(self, settings, variant):
        """
            Update banner
            @param settings as Gio.Settings
            @param value as GLib.Variant
        """
        self.__set_artwork()

    def _on_label_button_release(self, eventbox, event):
        """
            Show artists information
            @param eventbox as Gtk.EventBox
            @param event as Gdk.Event
        """
        if len(self.__artist_ids) == 1:
            from lollypop.pop_information import InformationPopover
            self.__pop_info = InformationPopover(True)
            self.__pop_info.set_relative_to(eventbox)
            self.__pop_info.populate(self.__artist_ids[0])
            self.__pop_info.show()

    def _on_play_clicked(self, *ignore):
        """
            Play artist albums
        """
        play_artists(self.__artist_ids, self.__genre_ids)

    def _on_add_clicked(self, *ignore):
        """
            Add artist albums
        """
        icon_name = self.__add_button.get_image().get_icon_name()[0]
        add = icon_name == "list-add-symbolic"
        add_artist_to_playback(self.__artist_ids, self.__genre_ids, add)

    def _on_menu_button_clicked(self, button):
        """
            Show album menu
            @param button as Gtk.Button
        """
        from lollypop.widgets_menu import MenuBuilder
        from lollypop.menu_artist import ArtistMenu
        from lollypop.menu_similars import SimilarsMenu
        from lollypop.menu_artwork import ArtistArtworkMenu
        menu = ArtistMenu(self.__artist_ids[0],
                          self.__storage_type,
                          self.view_type,
                          App().window.folded)
        menu_widget = MenuBuilder(menu, False)
        menu_widget.show()
        menu_ext = SimilarsMenu(self.__artist_ids[0])
        menu_ext.show()
        menu_widget.add_widget(menu_ext)
        if App().window.folded:
            menu_ext2 = ArtistArtworkMenu(self.__artist_ids[0],
                                          self.view_type,
                                          True)
            menu_ext2.connect("hidden", self.__close_artwork_menu)
            menu_ext2.show()
            menu_widget.add_widget(menu_ext2)
        popup_widget(menu_widget, button, None, None, button)

    def _on_badge_button_release(self, eventbox, event):
        """
            Show artist artwork manager
            @param eventbox as Gtk.EventBox
            @param event as Gdk.Event
        """
        from lollypop.widgets_menu import MenuBuilder
        from lollypop.menu_artwork import ArtistArtworkMenu
        menu = Gio.Menu()
        if App().window.folded:
            from lollypop.menu_header import ArtistMenuHeader
            menu.append_item(ArtistMenuHeader(self.__artist_ids[0]))
        menu_widget = MenuBuilder(menu, False)
        menu_widget.show()
        menu_ext = ArtistArtworkMenu(self.__artist_ids[0],
                                     self.view_type,
                                     False)
        menu_ext.connect("hidden", self.__close_artwork_menu)
        menu_ext.show()
        menu_widget.add_widget(menu_ext)
        self.__artwork_popup = popup_widget(menu_widget,
                                            eventbox,
                                            None, None, None)

    def _on_artist_artwork_changed(self, art, prefix):
        """
            Update artwork if needed
            @param art as Art
            @param prefix as str
        """
        if len(self.__artist_ids) == 1:
            artist = App().artists.get_name(self.__artist_ids[0])
            if prefix == artist:
                self.__set_artwork()
                self.__set_internal_size()

    def _on_playback_changed(self, *ignore):
        """
            Update add button
        """
        self.__update_add_button()

#######################
# PRIVATE             #
#######################
    def __close_artwork_menu(self, action, variant):
        if App().window.folded:
            App().window.container.go_back()
        else:
            self.__artwork_popup.destroy()

    def __set_artwork(self):
        """
            Set artwork
        """
        if App().settings.get_value("artist-artwork") and App().animations:
            artist = App().artists.get_name(choice(self.__artist_ids))
            App().art_helper.set_artist_artwork(
                                        artist,
                                        # +100 to prevent resize lag
                                        self.width + 100,
                                        self.height,
                                        self.get_scale_factor(),
                                        ArtBehaviour.BLUR_HARD |
                                        ArtBehaviour.DARKER,
                                        self._on_artwork)
        else:
            self._artwork.get_style_context().add_class("default-banner")
        if self.width < ArtSize.BANNER * 3:
            if self.__widget.get_child_at(1, 0) == self.__labels:
                self.__widget.remove(self.__labels)
                self.__widget.attach(self.__labels, 2, 2, 3, 1)
        elif self.__widget.get_child_at(2, 2) == self.__labels:
            self.__widget.remove(self.__labels)
            self.__widget.attach(self.__labels, 1, 0, 1, 3)

    def __set_badge_artwork(self, art_size):
        """
            Set artist artwork on badge
            @param art_size as int
        """
        if self.__show_artwork and\
                App().settings.get_value("artist-artwork"):
            artist = App().artists.get_name(self.__artist_ids[0])
            App().art_helper.set_artist_artwork(
                                        artist,
                                        art_size,
                                        art_size,
                                        self.get_scale_factor(),
                                        ArtBehaviour.ROUNDED |
                                        ArtBehaviour.CROP_SQUARE |
                                        ArtBehaviour.CACHE,
                                        self.__on_badge_artist_artwork,
                                        art_size)
        else:
            self.__badge_artwork.hide()

    def __set_internal_size(self):
        """
            Set content size based on current width
        """
        # Text size
        title_context = self.__title_label.get_style_context()
        for c in title_context.list_classes():
            title_context.remove_class(c)
        if App().window.folded:
            art_size = ArtSize.MEDIUM
            cls = "text-large"
        else:
            art_size = ArtSize.BANNER
            cls = "text-x-large"
        self.__title_label.get_style_context().add_class(cls)
        self.__set_badge_artwork(art_size)

    def __update_add_button(self):
        """
            Set image as +/-
        """
        album_ids = App().albums.get_ids(self.__genre_ids, self.__artist_ids,
                                         self.__storage_type, False)
        add = set(App().player.album_ids) & set(album_ids) != set(album_ids)
        if add:
            # Translators: artist context
            self.__add_button.set_tooltip_text(_("Add to current playlist"))
            self.__add_button.get_image().set_from_icon_name(
                "list-add-symbolic",
                Gtk.IconSize.BUTTON)
        else:
            # Translators: artist context
            self.__add_button.set_tooltip_text(
                _("Remove from current playlist"))
            self.__add_button.get_image().set_from_icon_name(
                "list-remove-symbolic",
                Gtk.IconSize.BUTTON)

    def __on_badge_artist_artwork(self, surface, art_size):
        """
            Set artist artwork on badge
            @param surface as cairo.Surface
            @param art_size as int
        """
        if surface is None:
            self.__badge_artwork.get_style_context().add_class("circle-icon")
            self.__badge_artwork.set_size_request(art_size, art_size)
            self.__badge_artwork.set_from_icon_name(
                                              "avatar-default-symbolic",
                                              Gtk.IconSize.DIALOG)
        else:
            self.__badge_artwork.get_style_context().remove_class(
                "circle-icon")
            self.__badge_artwork.set_from_surface(surface)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (C) 2010 Jonathan Matthew (replay gain code)
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gst

from lollypop.define import App, ReplayGain
from lollypop.logger import Logger


class PluginsPlayer:
    """
        Replay gain player
    """

    __EQUALIZER = [55, 77, 110, 156, 220, 311, 440, 622,
                   880, 1200, 1800, 2500, 3500, 5000, 7000,
                   10000, 14000, 20000]

    def __init__(self, playbin):
        """
            Init playbin
            @param playbin as Gst.bin
        """
        self.__equalizer = None
        self.__playbin = playbin
        self.build_audiofilter()

    def build_audiofilter(self):
        """
            Build audio filter
            audioconvert ! (rgvolume ! rglimiter ! audioconvert) !
            (equalizer) ! volume ! audioconvert ! audiosink
        """
        try:
            audiobin = Gst.ElementFactory.make("bin", None)
            audioconvert_in = Gst.ElementFactory.make("audioconvert", None)
            audiobin.add(audioconvert_in)
            # Replay gain
            replay_gain = App().settings.get_enum(
                "replay-gain") != ReplayGain.NONE
            if replay_gain:
                rgvolume = Gst.ElementFactory.make("rgvolume", None)
                rglimiter = Gst.ElementFactory.make("rglimiter", None)
                audioconvert_rg = Gst.ElementFactory.make("audioconvert", None)
                audiobin.add(rgvolume)
                audiobin.add(rglimiter)
                audiobin.add(audioconvert_rg)
                audioconvert_in.link(rgvolume)
                rgvolume.link(rglimiter)
                rglimiter.link(audioconvert_rg)
                if replay_gain == ReplayGain.ALBUM:
                    rgvolume.props.album_mode = 1
                else:
                    rgvolume.props.album_mode = 0
                rgvolume.props.pre_amp = App().settings.get_value(
                    "replay-gain-db").get_double()
                rglimiter.props.enabled = App().settings.get_value(
                    "replay-gain-limiter")

            # Internal volume manager
            self.volume = Gst.ElementFactory.make("volume", None)
            self.volume.props.volume = 1.0
            audiobin.add(self.volume)
            if replay_gain:
                audioconvert_rg.link(self.volume)
            else:
                audioconvert_in.link(self.volume)

            audiosink = Gst.ElementFactory.make("autoaudiosink", None)
            audiobin.add(audiosink)

            # Equalizer
            self.__equalizer = None
            if App().settings.get_value("equalizer-enabled"):
                self.__equalizer = Gst.ElementFactory.make("equalizer-nbands",
                                                           None)
# GStreamer's GstIirEqualizerNBands sets up shelve filters for the first and
# last bands as corner cases. That was causing the "inverted slider" bug.
# As a workaround, we create two dummy bands at both ends of the spectrum.
# This causes the actual first and last adjustable bands to be
# implemented using band-pass filters.
# Comment from Clementine source code, thanks Clementine devs!
                self.__equalizer.set_property("num-bands", 18 + 2)
                for idx in [0, 19]:
                    band = self.__equalizer.get_child_by_index(idx)
                    band.set_property("freq", 0)
                    band.set_property("bandwidth", 0)
                    band.set_property("gain", 0)
                # Setup bandwidth, thanks again to Clementine project
                last_band_freq = 0
                for idx in range(1, 19):
                    band = self.__equalizer.get_child_by_index(idx)
                    freq = self.__EQUALIZER[idx - 1]
                    bandwidth = freq - last_band_freq
                    last_band_freq = freq
                    band.set_property("bandwidth", bandwidth)
                audiobin.add(self.__equalizer)
                self.volume.link(self.__equalizer)
                self.__equalizer.link(audiosink)
            else:
                self.volume.link(audiosink)

            audiobin.add_pad(Gst.GhostPad.new(
                "sink",
                audioconvert_in.get_static_pad("sink")))
            self.__playbin.set_property("audio-sink", audiobin)
            if self.__equalizer is not None:
                self.update_equalizer()
        except Exception as e:
            Logger.error("PluginsPlayer::init():", e)

    def update_equalizer(self):
        """
            Update equalizer based on current settings
        """
        i = 0
        for value in App().settings.get_value("equalizer"):
            self.set_equalizer(i, value)
            i += 1

    def set_equalizer(self, index, value):
        """
            Set band value
            @param index as int
            @param value as int
        """
        try:
            if self.__equalizer is not None:
                band = self.__equalizer.get_child_by_index(index + 1)
                band.set_property("freq", self.__EQUALIZER[index])
                band.set_property("gain", value)
        except Exception as e:
            Logger.error("PluginsPlayer::set_equalizer():", e)

#######################
# PRIVATE             #
#######################



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GObject

from lollypop.helper_gestures import GesturesHelper
from lollypop.utils import set_cursor_type


class LabelWidget(Gtk.EventBox):
    """
        A clickable label
    """

    __gsignals__ = {
        "clicked": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    __LABEL_ATTRS = ["set_label", "set_markup", "set_xalign", "set_ellipsize",
                     "get_style_context", "set_justify"]

    def __init__(self):
        """
            Init label
        """
        Gtk.EventBox.__init__(self)
        self.__label = Gtk.Label.new()
        self.__label.show()
        self.add(self.__label)
        self.__gesture = GesturesHelper(
            self, primary_press_callback=self._on_press)
        self.connect("realize", set_cursor_type)

    def __getattr__(self, attr):
        """
            Get wanted attr
            @param attr as str
        """
        if attr in self.__LABEL_ATTRS:
            return getattr(self.__label, attr)
        else:
            return getattr(Gtk.EventBox, attr)

#######################
# PROTECTED           #
#######################
    def _on_press(self, x, y, event):
        self.emit("clicked")



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.define import App, Type
from lollypop.widgets_albums_rounded import RoundedAlbumsWidget


class AlbumsGenreWidget(RoundedAlbumsWidget):
    """
        Genre widget showing cover for 4 albums
    """

    def __init__(self, genre_id, storage_type, view_type, font_height):
        """
            Init widget
            @param Genre as [int]
            @param view_type as ViewType
            @param font_height as int
        """
        self.__storage_type = storage_type
        name = sortname = App().genres.get_name(genre_id)
        RoundedAlbumsWidget.__init__(self, genre_id, name, sortname,
                                     view_type, font_height)
        self._genre = Type.GENRES

    def populate(self):
        """
            Populate widget content
        """
        if self._artwork is None:
            RoundedAlbumsWidget.populate(self)
        else:
            self.set_artwork()

    @property
    def artwork_name(self):
        """
            Get artwork name
            return str
        """
        return "genre_" + self.name

#######################
# PROTECTED           #
#######################
    def _get_album_ids(self):
        """
            Get album ids
            @return [int]
        """
        return App().albums.get_ids([self._data], [],
                                    self.__storage_type, True)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib

import json
from locale import getdefaultlocale

from lollypop.define import App, AUDIODB_CLIENT_ID
from lollypop.utils import get_network_available
from lollypop.logger import Logger


class InformationDownloader:
    """
        Download info from the web
    """

    def __init__(self):
        """
            Init info downloader
        """
        pass

    def get_information(self, artist, callback, *args):
        """
            Get artist information
            @param artist as str
            @param callback as function
        """
        if not get_network_available("DATA"):
            callback(None, *args)
            return
        App().task_helper.run(self.__get_information, artist, callback, *args)

#######################
# PROTECTED           #
#######################
    def _get_audiodb_artist_info(self, artist):
        """
            Get artist info from audiodb
            @param artist as str
            @return info as bytes
        """
        if not get_network_available("AUDIODB"):
            return None
        try:
            artist = GLib.uri_escape_string(artist, None, True)
            uri = "https://theaudiodb.com/api/v1/json/"
            uri += "%s/search.php?s=%s" % (AUDIODB_CLIENT_ID, artist)
            (status, data) = App().task_helper.load_uri_content_sync(uri, None)
            if status:
                decode = json.loads(data.decode("utf-8"))
                language = getdefaultlocale()[0][-2:]
                for item in decode["artists"]:
                    for key in ["strBiography%s" % language,
                                "strBiographyEN"]:
                        info = item[key]
                        if info is not None:
                            return info.encode("utf-8")
        except Exception as e:
            Logger.error("InfoDownloader::_get_audiodb_artist_info: %s, %s" %
                         (e, artist))
        return None

    def _get_lastfm_artist_info(self, artist):
        """
            Get artist info from audiodb
            @param artist as str
            @return info as bytes
        """
        if get_network_available("LASTFM"):
            from lollypop.helper_web_lastfm import LastFMWebHelper
            helper = LastFMWebHelper()
            return helper.get_artist_bio(artist)
        return None

#######################
# PRIVATE             #
#######################
    def __get_information(self, artist, callback, *args):
        """
            Get information for artist
            @param artist as str
            @param callback as function
        """
        content = None
        try:
            # Try from Wikipedia first
            if get_network_available("WIKIPEDIA"):
                from lollypop.helper_web_wikipedia import WikipediaHelper
                wikipedia = WikipediaHelper()
                content = wikipedia.get_content_for_term(artist)
            if content is None:
                for method in [self._get_audiodb_artist_info,
                               self._get_lastfm_artist_info]:
                    content = method(artist)
                    if content is not None:
                        break
            callback(content, *args)
        except Exception as e:
            Logger.info("InfoDownloader::__get_information(): %s" % e)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GdkPixbuf, Gdk

from hashlib import md5

from lollypop.artwork_manager import ArtworkManager
from lollypop.logger import Logger
from lollypop.define import CACHE_PATH, ALBUMS_WEB_PATH, ALBUMS_PATH
from lollypop.define import ARTISTS_PATH, TimeStamp
from lollypop.utils import emit_signal
from lollypop.utils_file import remove_oldest, create_dir


class Artwork(ArtworkManager):
    """
        Artwork manager
    """

    def __init__(self):
        """
            Init artwork manager
        """
        ArtworkManager.__init__(self)
        create_dir(CACHE_PATH)

    def add_to_cache(self, name, surface, prefix, scale_factor):
        """
            Add artwork to cache
            @param name as str
            @param surface as cairo.Surface
            @param prefix as str
            @param scale_factor as int
            @thread safe
        """
        try:
            encoded = md5(name.encode("utf-8")).hexdigest()
            width = surface.get_width() * scale_factor
            height = surface.get_height() * scale_factor
            cache_path = "%s/@%s@%s_%s_%s" % (CACHE_PATH,
                                              prefix,
                                              encoded,
                                              width, height)
            pixbuf = Gdk.pixbuf_get_from_surface(surface, 0, 0, width, height)
            self.save_pixbuf(pixbuf, self.add_extension(cache_path))
        except Exception as e:
            Logger.error("Art::add_artwork_to_cache(): %s" % e)

    def remove_from_cache(self, name, prefix):
        """
            Remove artwork from cache
            @param name as str
            @param prefix as str
        """
        try:
            from glob import glob
            encoded = md5(name.encode("utf-8")).hexdigest()
            search = "%s/@%s@%s_*.%s" % (CACHE_PATH,
                                         prefix,
                                         encoded,
                                         self.extension_str)
            pathes = glob(search)
            for path in pathes:
                f = Gio.File.new_for_path(path)
                f.delete(None)
            emit_signal(self, "artwork-cleared", name, prefix)
        except Exception as e:
            Logger.error("Art::remove_artwork_from_cache(): %s" % e)

    def get_from_cache(self, name, prefix, width, height):
        """
            Get artwork from cache
            @param name as str
            @param prefix as str
            @param width as int
            @param height as int
            @return GdkPixbuf.Pixbuf
        """
        try:
            encoded = md5(name.encode("utf-8")).hexdigest()
            cache_path = "%s/@%s@%s_%s_%s" % (CACHE_PATH,
                                              prefix,
                                              encoded,
                                              width, height)
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(
                self.add_extension(cache_path))
            return pixbuf
        except Exception as e:
            Logger.warning("Art::get_artwork_from_cache(): %s" % e)
            return None

    def exists_in_cache(self, name, prefix, width, height):
        """
            True if artwork exists in cache
            @param name as str
            @param prefix as str
            @param width as int
            @param height as int
            @return bool
        """
        encoded = md5(name.encode("utf-8")).hexdigest()
        cache_path = "%s/@%s@%s_%s_%s" % (CACHE_PATH,
                                          prefix,
                                          encoded,
                                          width, height)
        f = Gio.File.new_for_path(self.add_extension(cache_path))
        return f.query_exists()

    def clean_artwork(self):
        """
            Remove old artwork from disk
        """
        try:
            remove_oldest(CACHE_PATH, TimeStamp.ONE_YEAR)
            remove_oldest(ARTISTS_PATH, TimeStamp.THREE_YEAR)
            remove_oldest(ALBUMS_PATH, TimeStamp.THREE_YEAR)
            remove_oldest(ALBUMS_WEB_PATH, TimeStamp.ONE_YEAR)
        except Exception as e:
            Logger.error("Art::clean_artwork(): %s", e)

    def clean_rounded(self):
        """
            Clean rounded artwork
        """
        try:
            from pathlib import Path
            extension = self.extension_str
            for p in Path(CACHE_PATH).glob("@ROUNDED*@*.%s" % extension):
                p.unlink()
        except Exception as e:
            Logger.error("Art::clean_all_cache(): %s", e)

    def clean_all_cache(self):
        """
            Remove all covers from cache
        """
        try:
            from pathlib import Path
            extension = self.extension_str
            for p in Path(CACHE_PATH).glob("*.%s" % extension):
                p.unlink()
        except Exception as e:
            Logger.error("Art::clean_all_cache(): %s", e)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio

from gettext import gettext as _
from time import time
from random import shuffle

from lollypop.view_flowbox import FlowBoxView
from lollypop.widgets_album_simple import AlbumSimpleWidget
from lollypop.define import App, Type, ViewType, ScanUpdate, StorageType
from lollypop.objects_album import Album
from lollypop.utils import get_icon_name, get_network_available, popup_widget
from lollypop.utils import get_title_for_genres_artists
from lollypop.utils import remove_static
from lollypop.utils_file import get_youtube_dl
from lollypop.utils_album import get_album_ids_for
from lollypop.helper_signals import SignalsHelper, signals_map


class AlbumsBoxView(FlowBoxView, SignalsHelper):
    """
        Show albums in a box
    """

    @signals_map
    def __init__(self, genre_ids, artist_ids, storage_type, view_type):
        """
            Init album view
            @param genre_ids as [int]
            @param artist_ids as [int]
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        FlowBoxView.__init__(self, storage_type, view_type)
        self.__time = time() + 10
        self._genre_ids = genre_ids
        self._artist_ids = artist_ids
        self._storage_type = storage_type
        self.__populate_wanted = True
        if genre_ids and genre_ids[0] < 0:
            if genre_ids[0] == Type.WEB:
                (youtube_dl, env) = get_youtube_dl()
                if youtube_dl is None:
                    self._empty_message = _("Missing youtube-dl command")
                    self.show_placeholder(True)
                    self.__populate_wanted = False
                elif not get_network_available("YOUTUBE"):
                    self._empty_message =\
                        _("Network unavailable or disabled in settings")
                    self.show_placeholder(True)
                    self.__populate_wanted = False
            self._empty_icon_name = get_icon_name(genre_ids[0])
        return [
            (App().scanner, "updated", "_on_collection_updated"),
            (App().player, "loading-changed", "_on_loading_changed"),
            (App().player, "current-changed", "_on_current_changed"),
            (App().album_art, "album-artwork-changed", "_on_artwork_changed")
        ]

    def populate(self, albums=[]):
        """
            Populate view for album ids
            Show artist_ids/genre_ids if empty
            @param albums as [Album]
        """
        def on_load(albums):
            if albums:
                FlowBoxView.populate(self, albums)
                self.show_placeholder(False)
            else:
                self.show_placeholder(True)

        def load():
            # No skipped albums for this views
            if self._genre_ids and self._genre_ids[0] in [Type.POPULARS,
                                                          Type.LITTLE,
                                                          Type.RANDOMS,
                                                          Type.RECENTS]:
                skipped = False
            else:
                skipped = True
            album_ids = get_album_ids_for(self._genre_ids, self._artist_ids,
                                          self.storage_type, skipped)
            albums = []
            for album_id in album_ids:
                album = Album(album_id, self._genre_ids,
                              self._artist_ids, True)
                album.set_storage_type(self.storage_type)
                albums.append(album)
            return albums

        if albums:
            FlowBoxView.populate(self, albums)
        elif self.__populate_wanted:
            App().task_helper.run(load, callback=(on_load,))

    def add_value(self, album):
        """
            Add a new album
            @param album as Album
        """
        self.show_placeholder(False)
        FlowBoxView.add_value_unsorted(self, album)

    def clear(self):
        """
            Clear view
        """
        for child in self._box.get_children():
            child.destroy()

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {"genre_ids": self._genre_ids,
                "artist_ids": self._artist_ids,
                "storage_type": self._storage_type,
                "view_type": self.view_type & ~ViewType.SMALL}

#######################
# PROTECTED           #
#######################
    def _get_child(self, value, position=-1):
        """
            Get a child for view
            @param value as object
            @param position as int
            @return row as SelectionListRow
        """
        if self.destroyed:
            return None
        widget = AlbumSimpleWidget(value,  self._genre_ids, self._artist_ids,
                                   self.view_type, self.font_height)
        self._box.insert(widget, position)
        widget.show()
        return widget

    def _get_menu_widget(self, child):
        """
            Get menu widget
            @param child as AlbumSimpleWidget
            @return Gtk.Widget
        """
        from lollypop.widgets_menu import MenuBuilder
        from lollypop.menu_objects import AlbumMenu
        menu = AlbumMenu(child.data, self.storage_type, self.view_type)
        return MenuBuilder(menu)

    def _on_current_changed(self, player):
        """
            Update children state
            @param player as Player
        """
        for child in self._box.get_children():
            child.set_selection()

    def _on_collection_updated(self, scanner, item, scan_update):
        """
            Handles changes in collection
            @param scanner as CollectionScanner
            @param item as CollectionItem
            @param scan_update as ScanUpdate
        """
        # On first update, ignore notifications for 10 seconds
        # Next, ignore notifications for 120 seconds
        if scan_update == ScanUpdate.ADDED and time() > self.__time:
            wanted = True
            for genre_id in item.genre_ids:
                genre_ids = remove_static(self._genre_ids)
                if genre_ids and genre_id not in genre_ids:
                    wanted = False
            for artist_id in item.artist_ids:
                artist_ids = remove_static(self._artist_ids)
                if artist_ids and artist_id not in artist_ids:
                    wanted = False
            if wanted:
                self.__time = time() + 120
                App().window.container.show_notification(
                    _("New albums available"),
                    [_("Refresh")],
                    [App().window.container.reload_view])
        elif scan_update == ScanUpdate.REMOVED:
            for child in self.children:
                if child.data.id == item.album_id:
                    child.destroy()
                    break

    def _on_artwork_changed(self, artwork, album_id):
        """
            Update children artwork if matching album id
            @param artwork as Artwork
            @param album_id as int
        """
        for child in self._box.get_children():
            if child.data.id == album_id:
                child.set_artwork()

    def _on_loading_changed(self, player, status, track):
        """
            Update row loading status
            @param player as Player
            @param status as bool
            @param track as Track
        """
        for child in self.children:
            if child.artwork is None:
                continue
            if child.data.id == track.album.id:
                context = child.artwork.get_style_context()
                if status:
                    context.add_class("load-animation")
                else:
                    context.remove_class("load-animation")
                break

    def _on_child_activated(self, flowbox, child):
        """
            Navigate into child
            @param flowbox as Gtk.FlowBox
            @param child as Gtk.FlowBoxChild
        """
        if child.artwork is None:
            return

        def show_album(status, child):
            child.artwork.get_style_context().remove_class("load-animation")
            App().window.container.show_view([Type.ALBUM], child.data,
                                             self.storage_type)

        if child.data.storage_type & StorageType.COLLECTION:
            App().window.container.show_view([Type.ALBUM], child.data)
        else:
            child.artwork.get_style_context().add_class("load-animation")
            cancellable = Gio.Cancellable.new()
            App().task_helper.run(child.data.load_tracks,
                                  cancellable,
                                  callback=(show_album, child))

    def _on_tertiary_press_gesture(self, x, y, event):
        """
            Play albums
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        child = self._box.get_child_at_pos(x, y)
        if child is None or child.artwork is None:
            return

        def play_album(status, child):
            child.artwork.get_style_context().remove_class("load-animation")
            App().player.play_album(child.data.clone(False))

        if child.data.storage_type & StorageType.COLLECTION:
            App().player.play_album(child.data.clone(False))
        else:
            child.artwork.get_style_context().add_class("load-animation")
            cancellable = Gio.Cancellable.new()
            App().task_helper.run(child.data.load_tracks,
                                  cancellable,
                                  callback=(play_album, child))


class AlbumsForGenresBoxView(AlbumsBoxView):
    """
        Show albums in a box for genres (static or not)
    """

    def __init__(self, genre_ids, artist_ids, storage_type, view_type):
        """
            Init album view
            @param genre_ids as [int]
            @param artist_ids as [int]
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        AlbumsBoxView.__init__(self, genre_ids, artist_ids, storage_type,
                               view_type | ViewType.OVERLAY)
        from lollypop.widgets_banner_flowbox import FlowboxBannerWidget
        self.__banner = FlowboxBannerWidget(genre_ids, artist_ids,
                                            view_type, True)
        self.__banner.show()
        self.__banner.connect("play-all", self.__on_banner_play_all)
        self.__banner.connect("show-menu", self.__on_banner_show_menu)
        self.add_widget(self._box, self.__banner)

#######################
# PROTECTED           #
#######################
    def _on_map(self, widget):
        """
            Set initial view state
            @param widget as GtK.Widget
        """
        AlbumsBoxView._on_map(self, widget)
        if self.view_type & ViewType.SCROLLED:
            self.scrolled.grab_focus()

#######################
# PRIVATE             #
#######################
    def __on_banner_play_all(self, banner, random):
        """
            Play all albums
            @param banner as AlbumsBannerWidget
            @param random as bool
        """
        albums = [c.data.clone(False) for c in self._box.get_children()]
        if not albums:
            return
        if random:
            shuffle(albums)
            App().player.play_album_for_albums(albums[0], albums)
        else:
            App().player.play_album_for_albums(albums[0], albums)

    def __on_banner_show_menu(self, banner, button):
        """
            Show contextual menu
            @param banner as AlbumsBannerWidget
            @param button as Gtk.Button
        """
        from lollypop.menu_objects import AlbumsMenu
        from lollypop.widgets_menu import MenuBuilder
        albums = []
        for child in self._box.get_children():
            if child.data.storage_type & StorageType.COLLECTION:
                albums.append(child.data)
        title = get_title_for_genres_artists(self._genre_ids, self._artist_ids)
        menu = AlbumsMenu(title, albums, self.view_type)
        menu_widget = MenuBuilder(menu)
        menu_widget.show()
        popup_widget(menu_widget, button, None, None, button)


class AlbumsForYearsBoxView(AlbumsForGenresBoxView):
    """
        Years album box
    """

    def __init__(self, genre_ids, artist_ids, storage_type, view_type):
        """
            Init view
            @param genre_ids as [int]
            @param artist_ids as [int]
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        AlbumsForGenresBoxView.__init__(self, genre_ids, artist_ids,
                                        storage_type, view_type)

    def populate(self):
        """
            Populate view
        """
        def on_load(items):
            FlowBoxView.populate(self, items)

        def get_album(album_id, disc_number, disc_name, album_year, year):
            album = Album(album_id, [Type.YEARS], [])
            if year != album_year:
                album.set_disc_number(disc_number)
            return album

        def load():
            albums = []
            for year in self._artist_ids:
                items = App().tracks.get_compilations_by_disc_for_year(
                    year, self.storage_type, True)
                items += App().tracks.get_albums_by_disc_for_year(
                    year, self.storage_type, True)
                albums += [get_album(item[0], item[1], item[2], item[3], year)
                           for item in items]
            return albums

        App().task_helper.run(load, callback=(on_load,))


class AlbumsDeviceBoxView(AlbumsBoxView):
    """
        Device album box
    """

    def __init__(self, index, view_type):
        """
            Init view
            @param index as int
            @param view_type as ViewType
            @param index as int
        """
        AlbumsBoxView.__init__(self, [], [], StorageType.COLLECTION, view_type)
        self.add_widget(self._box)
        self.__index = index

    def populate(self):
        """
            Populate view
        """
        def on_load(items):
            FlowBoxView.populate(self, items)

        def load():
            album_ids = App().albums.get_synced_ids(0)
            album_ids += App().albums.get_synced_ids(self.__index)
            return [Album(album_id) for album_id in album_ids]

        App().task_helper.run(load, callback=(on_load,))

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {"index": self.__index,
                "view_type": self.view_type & ~ViewType.SMALL}

#######################
# PROTECTED           #
#######################
    def _on_map(self, widget):
        """
            Set initial view state
            @param widget as GtK.Widget
        """
        AlbumsBoxView._on_map(self, widget)
        if self.view_type & ViewType.SCROLLED:
            self.scrolled.grab_focus()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

from lollypop.logger import Logger
from lollypop.helper_task import TaskHelper
from lollypop.utils import escape, get_network_available
from lollypop.utils_file import create_dir
from lollypop.define import LYRICS_PATH


class LyricsHelper:
    """
        Sync lyrics helper
    """

    def __init__(self):
        """
            Init helper
        """
        self.__timestamps = {}
        self.__cancellable = Gio.Cancellable.new()
        create_dir(LYRICS_PATH)

    def load(self, track):
        """
            Load lyrics for track
            @param track as Track
        """
        self.__track = track
        self.__timestamps = {}
        uri_no_ext = ".".join(track.uri.split(".")[:-1])
        self.__lrc_file = Gio.File.new_for_uri(uri_no_ext + ".lrc")
        if self.__lrc_file.query_exists():
            self.__get_timestamps()
        else:
            from lollypop.tagreader import Discoverer, TagReader
            discoverer = Discoverer()
            tagreader = TagReader()
            try:
                info = discoverer.get_info(track.uri)
            except:
                info = None
            if info is not None:
                tags = info.get_tags()
                for (lyrics, timestamp) in tagreader.get_synced_lyrics(tags):
                    if timestamp in self.__timestamps.keys():
                        self.__timestamps[timestamp] += "\n%s" % lyrics
                    else:
                        self.__timestamps[timestamp] = lyrics

    def get_lyrics_for_timestamp(self, timestamp):
        """
            Get lyrics for timestamp
            @param timestamp as int
            @return ([str], str, [str])
        """
        previous = []
        next = []
        current = None
        for key in self.__timestamps.keys():
            if timestamp > key and current is None:
                if len(previous) == 5:
                    previous.pop(0)
                previous.append(self.__timestamps[key])
            elif current is None and previous:
                current = previous.pop(-1)
                next.append(self.__timestamps[key])
            elif timestamp < key and len(next) != 5:
                next.append(self.__timestamps[key])
            else:
                break
        if current is None:
            current = ""
        return (previous, [" ", current, " "], next)

    def get_lyrics_from_web(self, track, callback, *args):
        """
            Get lyrics from web for track
            @param track as Track
            @param callback as function
        """
        self.__cancellable = Gio.Cancellable.new()
        methods = []
        if get_network_available("BING"):
            methods.append(self.__download_bing_lyrics)
        if get_network_available("GENIUS"):
            methods.append(self.__download_genius_lyrics)
        if get_network_available("METROLYRICS"):
            methods.append(self.__download_metro_lyrics)
        if methods:
            self.__get_lyrics_from_web(track, methods, callback, *args)
        else:
            callback(None, *args)

    def cancel(self):
        """
            Cancel current loading
        """
        self.__cancellable.cancel()

    @property
    def available(self):
        """
            True if lyrics available
            @return bool
        """
        return len(self.__timestamps.keys()) != 0

############
# PRIVATE  #
############
    def __str_to_timestamp(self, srt_timestamp):
        """
            Convert timestamp to time
            @timestamp as str [00:00.00]
            @return int
        """
        timestamp = int(srt_timestamp.split(".")[-1])
        seconds = int(srt_timestamp.split(".")[-2].split(":")[-1])
        minutes = int(srt_timestamp.split(".")[-2].split(":")[0])
        timestamp += seconds * 1000
        timestamp += minutes * 60000
        return timestamp

    def __get_timestamps(self):
        """
            Get timestamps from file
        """
        try:
            status = False
            if self.__lrc_file.query_exists():
                (status, content, tag) = self.__lrc_file.load_contents()
            if status:
                data = content.decode("utf-8").split("\n")
                for line in data:
                    if line.find("length") != -1:
                        continue
                    try:
                        str_timestamp = line.split("]")[0].split("[")[1]
                        timestamp = self.__str_to_timestamp(str_timestamp)
                        lyrics = " ".join(line.split("]")[1:])
                        if timestamp in self.__timestamps.keys():
                            self.__timestamps[timestamp] += "\n%s" % lyrics
                        else:
                            self.__timestamps[timestamp] = lyrics
                    except:
                        continue
        except Exception as e:
            Logger.error("SyncLyricsHelper::__get_timestamps(): %s", e)

    def __get_lyrics_from_web(self, track, methods, callback, *args):
        """
            Get lyrics from web for track
            @param track as Track
            @param methods as []
            @param callback as function
        """
        if methods:
            method = methods.pop(0)
            method(track, methods, callback, *args)
        else:
            callback("", *args)

    def __get_title(self, track, escape=False):
        """
            Get track title for lyrics
            @param track as Track
            @param escape as bool
            @return str
        """
        # Update lyrics
        title = track.name
        if escape:
            return GLib.uri_escape_string(title, None, False)
        else:
            return title

    def __get_artist(self, track, escape=False):
        """
            Get track artist for lyrics
            @param track as Track
            @param escape as bool
            @return str
        """
        if track.artists:
            artist = track.artists[0]
        elif track.album_artists:
            artist = track.album_artists[0]
        else:
            artist = ""
        if escape:
            return GLib.uri_escape_string(artist, None, False)
        else:
            return artist

    def __download_bing_lyrics(self, track, methods, callback, *args):
        """
            Downloas lyrics from metro lyrics
            @param track as Track
            @param methods as []
            @param callback as function
        """
        title = self.__get_title(track, False)
        artist = self.__get_artist(track, False).lower()
        string = "%s %s" % (artist, title)
        uri = "https://www.bing.com/search?q=%s" % string
        helper = TaskHelper()
        helper.load_uri_content(uri,
                                self.__cancellable,
                                self.__on_lyrics_downloaded,
                                "b_heroLyrics",
                                "\n",
                                track,
                                methods,
                                callback,
                                *args)

    def __download_metro_lyrics(self, track, methods, callback, *args):
        """
            Downloas lyrics from metro lyrics
            @param track as Track
            @param methods as []
            @param callback as function
        """
        title = self.__get_title(track, False)
        artist = self.__get_artist(track, False).lower()
        string = "%s-lyrics-%s" % (title, artist)
        uri = "https://www.metrolyrics.com/%s.html" % string.replace(" ", "-")
        helper = TaskHelper()
        helper.load_uri_content(uri,
                                self.__cancellable,
                                self.__on_lyrics_downloaded,
                                "lyrics-body",
                                "",
                                track,
                                methods,
                                callback,
                                *args)

    def __download_genius_lyrics(self, track, methods, callback, *args):
        """
            Download lyrics from genius
            @param track as Track
            @param methods as []
            @param callback as function
        """
        title = self.__get_title(track, False)
        artist = self.__get_artist(track, False)
        string = escape("%s %s" % (artist, title),
                        ignore=["_", "-", " ", ".", "/"]).replace(".", "")
        string = string.replace("/", " ")
        uri = "https://genius.com/%s-lyrics" % string.replace(" ", "-")
        helper = TaskHelper()
        helper.load_uri_content(uri,
                                self.__cancellable,
                                self.__on_lyrics_downloaded,
                                "Lyrics__Container-sc-1ynbvzw-6",
                                "\n",
                                track,
                                methods,
                                callback,
                                *args)

    def __on_lyrics_downloaded(self, uri, status, data, cls, separator,
                               track, methods, callback, *args):
        """
            Search lyrics and pass to callback
            @param uri as str
            @param status as bool
            @param data as bytes
            @param cls as str
            @param separator as str
            @param track as Track
            @param methods as []
            @param callback as function
        """
        Logger.debug("LyricsHelper::__on_lyrics_downloaded(): %s", uri)
        if status:
            try:
                from bs4 import BeautifulSoup
                soup = BeautifulSoup(data, 'html.parser')
                lyrics = soup.find_all(
                    "div", class_=cls)[0].get_text(separator=separator)
                callback(lyrics, *args)
                return
            except Exception as e:
                Logger.warning("LyricsHelper::__on_lyrics_downloaded(): %s", e)
        self.__get_lyrics_from_web(track, methods, callback, *args)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Handy

from lollypop.define import App, Size
from lollypop.toolbar_playback import ToolbarPlayback
from lollypop.objects_track import Track
from lollypop.toolbar_info import ToolbarInfo
from lollypop.toolbar_title import ToolbarTitle
from lollypop.toolbar_end import ToolbarEnd
from lollypop.helper_size_allocation import SizeAllocationHelper
from lollypop.helper_signals import SignalsHelper, signals


class Toolbar(Handy.HeaderBar, SizeAllocationHelper, SignalsHelper):
    """
        Lollypop toolbar
    """

    @signals
    def __init__(self, window):
        """
            Init toolbar
            @param window as Window
        """
        Handy.HeaderBar.__init__(self)
        SizeAllocationHelper.__init__(self)
        self.__width = Size.MINI
        self.__timeout_id = None
        self.set_title("Lollypop")
        self.__toolbar_playback = ToolbarPlayback(window)
        self.__toolbar_playback.show()
        self.__toolbar_info = ToolbarInfo()
        self.__toolbar_info.show()
        self.__toolbar_title = ToolbarTitle()
        self.__toolbar_end = ToolbarEnd(window)
        self.__toolbar_end.show()
        self.pack_start(self.__toolbar_playback)
        self.pack_start(self.__toolbar_info)
        self.set_custom_title(self.__toolbar_title)
        self.pack_end(self.__toolbar_end)
        return [
            (App().player, "current-changed", "_on_current_changed"),
            (App().window.container.widget, "notify::folded",
             "_on_container_folded")
        ]

    def do_get_preferred_width(self):
        """
            Allow window resize
            @return (int, int)
        """
        width = max(Size.PHONE, self.__width)
        return (Size.PHONE, width)

    def hide_info_and_buttons(self, folded):
        """
            Hide information and buttons
            @param folded as bool
        """
        def show_children():
            self.__timeout_id = None
            self.__toolbar_info.show_children()

        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)
            self.__timeout_id = None

        if folded:
            self.__toolbar_playback.player_buttons.hide()
            self.__toolbar_title.hide()
            self.__toolbar_info.hide_children()
        else:
            if App().player.current_track.id is not None:
                self.__toolbar_title.show()
            # If user double click headerbar to maximize window
            # We do not want info bar to receive click signal
            self.__timeout_id = GLib.timeout_add(200, show_children)
            self.__toolbar_playback.player_buttons.show()

    @property
    def end(self):
        """
            Return end toolbar
            @return ToolbarEnd
        """
        return self.__toolbar_end

    @property
    def info(self):
        """
            Return info toolbar
            @return ToolbarInfo
        """
        return self.__toolbar_info

    @property
    def title(self):
        """
            Return title toolbar
            @return ToolbarTitle
        """
        return self.__toolbar_title

    @property
    def playback(self):
        """
            Return playback toolbar
            @return ToolbarPlayback
        """
        return self.__toolbar_playback

#######################
# PROTECTED           #
#######################
    def _handle_width_allocate(self, allocation):
        """
            Update artwork
            @param allocation as Gtk.Allocation
        """
        if SizeAllocationHelper._handle_width_allocate(self, allocation):
            width = self.__toolbar_playback.get_preferred_width()[1]
            width += self.__toolbar_end.get_preferred_width()[1]
            available = allocation.width - width
            if allocation.width < Size.BIG:
                title_width = available / 3
            else:
                title_width = available / 2.5
            self.__toolbar_title.set_width(title_width)
            self.__toolbar_info.set_width((available - title_width) / 2)

    def _on_current_changed(self, player):
        """
            Update buttons and progress bar
            @param player as Player
        """
        if player.current_track.id is not None and\
                not App().window.folded:
            if isinstance(player.current_track, Track):
                self.__toolbar_title.show()
            else:
                self.__toolbar_title.hide()
            self.__toolbar_info.show_children()
        else:
            self.__toolbar_title.hide()
            self.__toolbar_info.hide_children()

    def _on_container_folded(self, leaflet, folded):
        """
            Update internal widgets
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        self.hide_info_and_buttons(App().window.folded)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from lollypop.widgets_player_progress import ProgressPlayerWidget


class ToolbarTitle(Gtk.Bin):
    """
        Title toolbar
    """

    def __init__(self):
        """
            Init toolbar
        """
        Gtk.Bin.__init__(self)
        progress_widget = ProgressPlayerWidget()
        progress_widget.show()
        self.add(progress_widget)

    def set_width(self, width):
        """
            Set Gtk.Scale progress width
            @param width as int
        """
        self.set_property("width_request", width)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Pango, GLib

from gettext import gettext as _

from lollypop.widgets_row_track import TrackRow
from lollypop.define import App, ViewType
from lollypop.view_tracks import TracksView
from lollypop.helper_signals import SignalsHelper, signals_map


class SearchTracksView(TracksView, SignalsHelper):
    """
        Responsive view showing searched tracks
    """

    @signals_map
    def __init__(self):
        """
            Init view
        """
        TracksView.__init__(self, ViewType.SEARCH)
        self.__track_ids = []
        self._add_disc_container(0)
        self._tracks_widget_left[0].show()
        self._tracks_widget_right[0].show()
        return [
            (App().window, "notify::folded", "_on_container_folded"),
        ]

    def append_row(self, track):
        """
            Append a track
            ONE COLUMN ONLY
            @param track as Track
            @param position as int
        """
        self._init()
        if track.id in self.__track_ids:
            return
        self.__track_ids.append(track.id)
        left_len = len(self._tracks_widget_left[0].get_children())
        right_len = len(self._tracks_widget_right[0].get_children())
        if left_len > right_len:
            self._add_tracks(self._tracks_widget_right[0], [track])
        else:
            self._add_tracks(self._tracks_widget_left[0], [track])

    def clear(self):
        """
            Clear and hide the view
        """
        self.__track_ids = []
        for child in self._tracks_widget_left[0].get_children() +\
                self._tracks_widget_right[0].get_children():
            child.destroy()
        GLib.idle_add(self.hide)

    @property
    def children(self):
        """
            Return all rows
            @return [Gtk.ListBoxRow]
        """
        return self._tracks_widget_left[0].get_children() +\
            self._tracks_widget_right[0].get_children()

#######################
# PROTECTED           #
#######################
    def _add_tracks(self, widget, tracks, position=0):
        """
            Add tracks to widget
            @param widget as Gtk.ListBox
            @param tracks as [Track]
        """
        for track in tracks:
            track.set_number(position + 1)
            row = TrackRow(track, [], self.view_type)
            row.show()
            widget.add(row)
            position += 1

    def _set_orientation(self, orientation):
        """
            Set columns orientation
            @param orientation as Gtk.Orientation
        """
        if not TracksView._set_orientation(self, orientation):
            return
        self._responsive_widget.insert_row(0)
        self.__label = Gtk.Label.new(_("Tracks"))
        self.__label.show()
        self.__label.set_ellipsize(Pango.EllipsizeMode.END)
        self.__label.get_style_context().add_class("dim-label")
        self.__update_label(App().window.folded)
        self.__label.set_hexpand(True)
        self.__label.set_property("halign", Gtk.Align.START)
        idx = 1
        # Vertical
        ##########################
        #  --------Label-------- #
        #  |     Column 1      | #
        #  |     Column 2      | #
        ##########################
        # Horizontal
        ###########################
        # ---------Label--------- #
        # | Column 1 | Column 2 | #
        ###########################
        if orientation == Gtk.Orientation.VERTICAL:
            self._responsive_widget.attach(self.__label, 0, 0, 1, 1)
            self._responsive_widget.attach(
                      self._tracks_widget_left[0],
                      0, idx, 2, 1)
            idx += 1
        else:
            self._responsive_widget.attach(self.__label, 0, 0, 2, 1)
            self._responsive_widget.attach(
                      self._tracks_widget_left[0],
                      0, idx, 1, 1)
        if not self.view_type & ViewType.SINGLE_COLUMN:
            if orientation == Gtk.Orientation.VERTICAL:
                self._responsive_widget.attach(
                           self._tracks_widget_right[0],
                           0, idx, 2, 1)
            else:
                self._responsive_widget.attach(
                           self._tracks_widget_right[0],
                           1, idx, 1, 1)
        idx += 1

    def _on_loading_changed(self, player, status, track):
        """
            Update row loading status
            @param player as Player
            @param status as bool
            @param track as Track
        """
        if track.is_web:
            TracksView._on_loading_changed(self, player, status, track)

    def _on_activated(self, widget, track):
        """
            Handle playback if album or pass signal
            @param widget as TracksWidget
            @param track as Track
        """
        track.album.set_tracks([track])
        App().player.add_album(track.album)
        App().player.load(track.album.get_track(track.id))

    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        self.__update_label(App().window.folded)

#######################
# PRIVATE             #
#######################
    def __update_label(self, is_adaptive):
        """
            Update label style based on current adaptive state
            @param is_adaptive as bool
        """
        style_context = self.__label.get_style_context()
        if is_adaptive:
            style_context.remove_class("text-x-large")
        else:
            style_context.add_class("text-x-large")



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (c) 2015 Jean-Philippe Braun <eon@patapon.info>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.logger import Logger
from lollypop.define import App
from lollypop.utils import emit_signal


class Base:
    """
        Base for album and track objects
    """

    def __init__(self, db):
        self.db = db

    def __dir__(self, *args, **kwargs):
        """
            Concatenate base class"s fields with child class"s fields
        """
        return super(Base, self).__dir__(*args, **kwargs) +\
            list(self.DEFAULTS.keys())

    def __getattr__(self, attr):
        # Lazy DB calls of attributes
        if attr in list(self.DEFAULTS.keys()):
            if self.id is None or self.id < 0:
                return self.DEFAULTS[attr]
            # Actual value of "attr_name" is stored in "_attr_name"
            attr_name = "_" + attr
            attr_value = getattr(self, attr_name)
            if attr_value is None:
                attr_value = getattr(self.db, "get_" + attr)(self.id)
                setattr(self, attr_name, attr_value)
            # Return default value if None
            if attr_value is None:
                return self.DEFAULTS[attr]
            else:
                return attr_value

    def reset(self, attr):
        """
            Reset attr
            @param attr as str
        """
        attr_name = "_" + attr
        attr_value = getattr(self.db, "get_" + attr)(self.id)
        setattr(self, attr_name, attr_value)

    def get_popularity(self):
        """
            Get popularity
            @return int between 0 and 5
        """
        if self.id is None:
            return 0

        popularity = 0
        avg_popularity = self.db.get_avg_popularity()
        if avg_popularity > 0:
            popularity = self.db.get_popularity(self.id)
        return popularity * 5 / avg_popularity + 0.5

    def set_popularity(self, new_rate):
        """
            Set popularity
            @param new_rate as int between 0 and 5
        """
        if self.id is None:
            return
        try:
            avg_popularity = self.db.get_avg_popularity()
            popularity = int((new_rate * avg_popularity / 5) + 0.5)
            best_popularity = self.db.get_higher_popularity()
            if new_rate == 5:
                popularity = (popularity + best_popularity) / 2
            self.db.set_popularity(self.id, popularity)
            self.reset("popularity")
        except Exception as e:
            Logger.error("Base::set_popularity(): %s" % e)

    def set_rate(self, rate):
        """
            Set rate
            @param rate as int between -1 and 5
        """
        self.db.set_rate(self.id, rate)
        self.reset("rate")
        emit_signal(App().player, "rate-changed", self.id, rate)
